import os
import sys
from six import iteritems

from . import build
from . import log
from . import path
from .arguments import parser as argparse
from .backends import list_backends
from .environment import Environment, EnvVersionError
from .platforms.target import platform_info
from .app_version import version

logger = log.getLogger(__name__)

description = """
bfg9000 ("build file generator") is a cross-platform build configuration system
with an emphasis on making it easy to define how to build your software. It
converts a Python-based build script into the appropriate files for your
underlying build system of choice.
"""

configure_desc = """
Generate the necessary build files to perform actual builds. If DIRECTORY is a
source directory (i.e. it contains a build.bfg file), the build files will be
created in the current directory. Otherwise, DIRECTORY is treated as the build
directory, and bfg9000 will look for a build.bfg file in the current directory.
"""

configureinto_desc = """
Generate the necessary build files to perform actual builds from a build.bfg
file in SRCDIR, and place them in BUILDDIR.
"""

refresh_desc = """
Regenerate an existing set of build files needed to perform actual builds. This
is run automatically if bfg9000 determines that the build files are out of
date.
"""

env_desc = """
Print the environment variables stored by this build configuration.
"""

e1m1_desc = """
You find yourself standing at Doom's gate.
"""


def handle_reload_exception(e, suggest_rerun=False):
    msg = 'Unable to reload environment'
    if str(e):
        msg += ': {}'.format(str(e))
    if suggest_rerun and isinstance(e, EnvVersionError):
        msg += '\n  Please re-run bfg9000 manually'
    logger.error(msg, exc_info=True)
    return 1


def environment_from_args(args):
    # Get the bin directory holding bfg's executables.
    bfgdir = path.abspath(sys.argv[0]).parent()

    backend = list_backends()[args.backend]
    env = Environment(
        bfgdir=bfgdir,
        backend=args.backend,
        backend_version=backend.version(),
        srcdir=args.srcdir,
        builddir=args.builddir,
    )

    return env, backend


def finalize_environment(env, args, extra_args=None):
    env.finalize(
        install_dirs={i: getattr(args, i.name) for i in path.InstallRoot},
        library_mode=(args.shared, args.static),
        extra_args=extra_args,
    )


def directory_pair(srcname, buildname):
    class DirectoryPair(argparse.Action):
        def __call__(self, parser, namespace, values, option_string=None):
            cwd = path.abspath('.')

            if build.is_srcdir(values):
                srcdir, builddir = values, cwd
            else:
                srcdir, builddir = cwd, values

            setattr(namespace, srcname, srcdir)
            setattr(namespace, buildname, builddir)

    return DirectoryPair


class ConfigureHelp(argparse.Action):
    def __init__(self, option_strings, dest=argparse.SUPPRESS,
                 default=argparse.SUPPRESS, help=None):
        argparse.Action.__init__(
            self, option_strings=option_strings, dest=dest, default=default,
            nargs=0, help=help
        )

    def __call__(self, parser, namespace, values, option_string=None):
        cwd = path.abspath('.')
        if not hasattr(namespace, 'srcdir') and build.is_srcdir(cwd):
            namespace.srcdir = cwd
            namespace.builddir = None

        if getattr(namespace, 'srcdir', None):
            env, backend = environment_from_args(namespace)
            parser = build.fill_user_help(env, parser)

        parser.print_help()
        parser.exit()


def add_generic_args(parser):
    parser.add_argument('--version', action='version',
                        version='%(prog)s ' + version)
    parser.add_argument('--debug', action='store_true', help=argparse.SUPPRESS)
    parser.add_argument('--color', metavar='WHEN',
                        choices=['always', 'never', 'auto'], default='auto',
                        help=('show colored output (one of: %(choices)s; ' +
                              'default: %(default)s)'))
    parser.add_argument('-c', action='store_const', const='always',
                        dest='color',
                        help=('show colored output (equivalent to ' +
                              '`--color=always`)'))
    parser.add_argument('--warn-once', action='store_true',
                        help='only emit a given warning once')


def add_configure_args(parser):
    backends = list_backends()

    parser.add_argument('-h', '--help', action=ConfigureHelp,
                        help='show this help message and exit')

    build = parser.add_argument_group('build arguments')
    build.add_argument('--backend', metavar='BACKEND',
                       choices=list(backends.keys()),
                       default=list(backends.keys())[0],
                       help=('build backend (one of %(choices)s; default: ' +
                             '%(default)s)'))
    build.add_argument('--toolchain', metavar='FILE',
                       type=argparse.File(must_exist=True),
                       help=('a file defining the toolchain to use for this ' +
                             'build'))
    build.add_argument('--shared', action='enable', default=True,
                       help='build shared libraries (default: enabled)')
    build.add_argument('--static', action='enable', default=False,
                       help='build static libraries (default: disabled)')

    common_path_help = 'installation path for {} (default: {{}})'
    path_help = {
        'prefix': 'installation prefix (default: {})',
        'exec_prefix': ('installation prefix for architecture-dependent ' +
                        'files (default: {})'),
        'bindir': common_path_help.format('executables'),
        'libdir': common_path_help.format('libraries'),
        'includedir': common_path_help.format('headers'),
    }

    install_dirs = platform_info().install_dirs
    install = parser.add_argument_group('installation arguments')
    for root in path.InstallRoot:
        name = '--' + root.name.replace('_', '-')
        help = path_help[root.name].format(repr(install_dirs[root]))
        install.add_argument(name, type=argparse.Directory(), metavar='PATH',
                             help=help)


def configure(parser, subparser, args, extra):
    if ( path.exists(args.builddir) and
         path.samefile(args.srcdir, args.builddir) ):
        subparser.error('source and build directories must be different')
    if not build.is_srcdir(args.srcdir):
        subparser.error('source directory must contain a {} file'
                        .format(build.bfgfile))
    if build.is_srcdir(args.builddir):
        subparser.error('build directory must not contain a {} file'
                        .format(build.bfgfile))

    if not path.exists(args.builddir):
        os.mkdir(args.builddir.string())

    try:
        env, backend = environment_from_args(args)
        if args.toolchain:
            build.load_toolchain(env, args.toolchain)
        finalize_environment(env, args, extra)
        env.save(args.builddir.string())

        build_inputs = build.configure_build(env)
        backend.write(env, build_inputs)
    except Exception as e:
        logger.exception(e)
        return 1


def refresh(parser, subparser, args, extra):
    if extra:
        subparser.error('unrecognized arguments: {}'.format(' '.join(extra)))

    if build.is_srcdir(args.builddir):
        subparser.error('build directory must not contain a {} file'
                        .format(build.bfgfile))

    try:
        env = Environment.load(args.builddir.string())
        if env.toolchain.path:
            build.load_toolchain(env, env.toolchain.path, reload=True)
        env.save(args.builddir.string())

        backend = list_backends()[env.backend]
        build_inputs = build.configure_build(env)
        backend.write(env, build_inputs)
    except Exception as e:
        return handle_reload_exception(e, suggest_rerun=True)


def env(parser, subparser, args, extra):
    if extra:
        subparser.error('unrecognized arguments: {}'.format(' '.join(extra)))

    if build.is_srcdir(args.builddir):
        subparser.error('build directory must not contain a {} file'
                        .format(build.bfgfile))

    try:
        env = Environment.load(args.builddir.string())

        for k, v in sorted(iteritems(env.variables)):
            if not args.unique or os.getenv(k) != v:
                print('{}={}'.format(k, v))
    except Exception as e:
        return handle_reload_exception(e)


def e1m1(parser, subparser, args, extra):  # pragma: no cover
    import e1m1
    try:
        e1m1.play(args.tempo, args.long)
    except Exception as e:
        logger.exception(e)
        return 1


def help(parser, subparser, args, extra):
    parser.parse_args(extra + ['--help'])


def main():
    parser = argparse.ArgumentParser(prog='bfg9000', description=description)
    subparsers = parser.add_subparsers(metavar='COMMAND')
    subparsers.required = True

    add_generic_args(parser)

    conf_p = subparsers.add_parser(
        'configure', description=configure_desc, add_help=False,
        help='create build files'
    )
    conf_p.set_defaults(func=configure, parser=conf_p)
    conf_p.add_argument(argparse.SUPPRESS,
                        action=directory_pair('srcdir', 'builddir'),
                        type=argparse.Directory(), metavar='DIRECTORY',
                        help='source or build directory')
    add_configure_args(conf_p)

    confinto_p = subparsers.add_parser(
        'configure-into', description=configureinto_desc, add_help=False,
        help='create build files in a chosen directory'
    )
    confinto_p.add_argument('srcdir', type=argparse.Directory(must_exist=True),
                            metavar='SRCDIR', help='source directory')
    confinto_p.add_argument('builddir', type=argparse.Directory(),
                            metavar='BUILDDIR', help='build directory')
    confinto_p.set_defaults(func=configure, parser=confinto_p)
    add_configure_args(confinto_p)

    refresh_p = subparsers.add_parser(
        'refresh', description=refresh_desc, help='regenerate build files'
    )
    refresh_p.set_defaults(func=refresh, parser=refresh_p)
    refresh_p.add_argument('builddir',
                           type=argparse.Directory(must_exist=True),
                           metavar='BUILDDIR', nargs='?', default='.',
                           help='build directory')

    env_p = subparsers.add_parser(
        'env', description=env_desc, help='print environment'
    )
    env_p.set_defaults(func=env, parser=env_p)
    env_p.add_argument('-u', '--unique', action='store_true',
                       help='only show variables that differ from the ' +
                       'current environment')
    env_p.add_argument('builddir', type=argparse.Directory(must_exist=True),
                       metavar='BUILDDIR', nargs='?', default='.',
                       help='build directory')

    e1m1_p = subparsers.add_parser('e1m1', description=e1m1_desc)
    e1m1_p.set_defaults(func=e1m1, parser=e1m1_p)
    # Windows gets glitchy if we play back too fast...
    tempo = 100 if platform_info().family == 'windows' else 120
    e1m1_p.add_argument('-t', '--tempo', metavar='BPM', type=int,
                        default=tempo,
                        help='playback speed (default:  %(default)s)')
    e1m1_p.add_argument('-L', '--long', action='store_true',
                        help='play for longer')

    help_p = subparsers.add_parser(
        'help', help='show this help message and exit', add_help=False
    )
    help_p.set_defaults(func=help, parser=help_p)

    args, extra = parser.parse_known_args()
    log.init(args.color, debug=args.debug, warn_once=args.warn_once)

    return args.func(parser, args.parser, args, extra)


def simple_main():
    parser = argparse.ArgumentParser(prog='9k', description=configure_desc,
                                     add_help=False)
    parser.add_argument(argparse.SUPPRESS,
                        action=directory_pair('srcdir', 'builddir'),
                        type=argparse.Directory(), metavar='DIRECTORY',
                        help='source or build directory')
    add_generic_args(parser)
    add_configure_args(parser)

    args, extra = parser.parse_known_args()
    log.init(args.color, debug=args.debug, warn_once=args.warn_once)

    return configure(parser, parser, args, extra)



from six import iteritems

from . import builtin
from ..build_inputs import build_input
from ..iterutils import default_sentinel


@build_input('project')
class ProjectInfo(object):
    def __init__(self, build_inputs, env):
        self.name = env.srcdir.basename()
        self.version = None
        self._options = {
            'intermediate_dirs': True,
            'lang': 'c',
        }

    def __getitem__(self, key):
        return self._options[key]

    def __setitem__(self, key, value):
        if key not in self._options:
            raise KeyError('unknown option {!r}'.format(key))
        self._options[key] = value


@builtin.function('build_inputs')
def project(build, name=default_sentinel, version=default_sentinel, **kwargs):
    info = build['project']
    if name is not default_sentinel:
        info.name = name
    if version is not default_sentinel:
        info.version = version

    for k, v in iteritems(kwargs):
        info[k] = v



import os
import re

from ... import path
from ... import shell
from .solution import Solution, UuidMap
from .syntax import *  # noqa
from ...versioning import Version


def version(env=os.environ):
    try:
        msbuild = shell.which(env.get('MSBUILD', ['msbuild', 'xbuild']), env)
        output = shell.execute(msbuild + ['/version'], stdout=shell.Mode.pipe,
                               stderr=shell.Mode.devnull)
        m = re.search(r'([\d\.]+)$', output)
        if m:
            return Version(m.group(1))
    except (IOError, OSError, shell.CalledProcessError):
        pass
    return None


priority = 1

_rule_handlers = {}


def rule_handler(*args):
    def decorator(fn):
        for i in args:
            _rule_handlers[i] = fn
        return fn
    return decorator


def write(env, build_inputs):
    uuids = UuidMap(env.builddir.append('.bfg_uuid').string())
    solution = Solution(uuids)

    for e in build_inputs.edges():
        _rule_handlers[type(e)](e, build_inputs, solution, env)

    # XXX: Handle default builds. Default builds go first in the solution. This
    # also means we'd need to support aliases so that we can have multiple
    # builds be the default.
    sln_file = path.Path(build_inputs['project'].name + '.sln')
    with open(sln_file.string(env.base_dirs), 'w') as out:
        solution.write(out)
    for p in solution:
        path.makedirs(p.path.parent().string(env.base_dirs), exist_ok=True)
        with open(p.path.string(env.base_dirs), 'w') as out:
            p.write(out)
    uuids.save()



import argparse
import subprocess

from . import tool
from .common import SimpleCommand
from .. import options as opts, shell
from ..exceptions import PackageResolutionError, PackageVersionError
from ..objutils import memoize
from ..packages import Package, PackageKind
from ..path import Path, Root
from ..versioning import check_version, Version


@tool('pkg_config')
class PkgConfig(SimpleCommand):
    _options = {
        'version': ['--modversion'],
        'cflags': ['--cflags'],
        'lib_dirs': ['--libs-only-L'],
        'ldflags': ['--libs-only-L', '--libs-only-other'],
        'ldlibs': ['--libs-only-l'],
    }

    def __init__(self, env):
        SimpleCommand.__init__(self, env, name='pkg_config',
                               env_var='PKG_CONFIG', default='pkg-config')

    def _call(self, cmd, name, type, static=False, msvc_syntax=False):
        result = cmd + [name] + self._options[type]
        if static:
            result.append('--static')
        if msvc_syntax:
            result.append('--msvc-syntax')
        return result


class PkgConfigPackage(Package):
    def __init__(self, name, format, specifier, kind, pkg_config):
        self._pkg_config = pkg_config

        try:
            version = Version(self._pkg_config.run(name, 'version').strip())
        except subprocess.CalledProcessError:
            raise PackageResolutionError("unable to find package '{}'"
                                         .format(name))

        check_version(version, specifier, name, PackageVersionError)
        self.version = version
        self.specifier = specifier
        self.static = kind == PackageKind.static
        Package.__init__(self, name, format)

    @memoize
    def _call(self, *args, **kwargs):
        return shell.split(self._pkg_config.run(*args, **kwargs).strip(),
                           type=opts.option_list)

    def compile_options(self, compiler):
        return self._call(self.name, 'cflags', self.static,
                          compiler.flavor == 'msvc')

    def link_options(self, linker):
        flags = self._call(self.name, 'ldflags', self.static,
                           linker.flavor == 'msvc')

        # XXX: How should we ensure that these libs are linked statically when
        # necessary?
        libs = self._call(self.name, 'ldlibs', self.static,
                          linker.flavor == 'msvc')
        libs = opts.option_list(opts.lib_literal(i) for i in libs)

        if linker.builder.object_format != 'elf' or self.static:
            return flags + libs

        # pkg-config packages don't generally include rpath information, so we
        # need to generate it ourselves.
        dir_args = self._call(self.name, 'lib_dirs', self.static,
                              linker.flavor == 'msvc',
                              env={'PKG_CONFIG_ALLOW_SYSTEM_LIBS': '1'})

        parser = argparse.ArgumentParser()
        parser.add_argument('-L', action='append', dest='lib_dirs')
        lib_dirs = parser.parse_known_args(dir_args)[0].lib_dirs or []
        rpaths = opts.option_list(opts.rpath_dir(Path(i, Root.absolute))
                                  for i in lib_dirs)

        return flags + libs + rpaths

    def __repr__(self):
        return '<PkgConfigPackage({!r}, {!r})>'.format(
            self.name, str(self.version)
        )


def resolve(env, name, format, version=None, kind=PackageKind.any):
    return PkgConfigPackage(name, format, version, kind,
                            env.tool('pkg_config'))



from . import tool
from .common import SimpleCommand


@tool('mkdir_p')
class MkdirP(SimpleCommand):
    def __init__(self, env):
        default = ('doppel -p' if env.host_platform.family == 'windows'
                   else 'mkdir -p')
        SimpleCommand.__init__(self, env, name='mkdir_p', env_var='MKDIR_P',
                               default=default)

    def _call(self, cmd, path):
        return cmd + [path]



import os.path

from . import *


class TestCommand(IntegrationTest):
    def __init__(self, *args, **kwargs):
        IntegrationTest.__init__(
            self, os.path.join(examples_dir, '08_commands'), *args, **kwargs
        )

    def test_hello(self):
        assertRegex(self, self.build('hello'), r'(?m)^\s*hello$')

    def test_world(self):
        assertRegex(self, self.build('world'), r'(?m)^\s*world$')

    def test_script(self):
        assertRegex(self, self.build('script'), r'(?m)^\s*hello, world!$')
        self.assertExists(output_file('file'))

    def test_alias(self):
        output = self.build('hello-world')
        assertRegex(self, output, r'(?m)^\s*hello$')
        assertRegex(self, output, r'(?m)^\s*world$')


@skip_if_backend('msbuild')
class TestRunExecutable(IntegrationTest):
    def __init__(self, *args, **kwargs):
        IntegrationTest.__init__(self, 'run_executable', *args, **kwargs)

    def test_env_run(self):
        self.assertExists(output_file('file.txt'))

    def test_cxx(self):
        assertRegex(self, self.build('cxx'), r'(?m)^\s*hello from c\+\+!$')

    def test_java(self):
        assertRegex(self, self.build('java'), r'(?m)^\s*hello from java!$')

    def test_java_classlist(self):
        assertRegex(self, self.build('java-classlist'),
                    r'(?m)^\s*hello from java!$')

    def test_python(self):
        assertRegex(self, self.build('python'), r'(?m)^\s*hello from python!$')



import os
import re
from itertools import chain
from six import string_types

from . import pkg_config
from .. import log, options as opts, safe_str, shell
from .ar import ArLinker
from .common import (BuildCommand, Builder, check_which, darwin_install_name,
                     library_macro)
from .ld import LdLinker
from ..builtins.copy_file import CopyFile
from ..exceptions import PackageResolutionError
from ..file_types import *
from ..iterutils import (default_sentinel, first, iterate, listify, uniques,
                         recursive_walk)
from ..languages import known_formats
from ..packages import CommonPackage, Framework, PackageKind
from ..path import abspath, exists, BasePath, InstallRoot, Path, Root
from ..platforms import parse_triplet
from ..versioning import detect_version, SpecifierSet

_optimize_flags = {
    opts.OptimizeValue.disable : '-O0',
    opts.OptimizeValue.size    : '-Osize',
    opts.OptimizeValue.speed   : '-O3',
    opts.OptimizeValue.linktime: '-flto',
}


class CcBuilder(Builder):
    def __init__(self, env, langinfo, command, version_output):
        brand, version, target_flags = self._parse_brand(env, command,
                                                         version_output)
        Builder.__init__(self, langinfo.name, brand, version)
        self.object_format = env.target_platform.object_format

        name = langinfo.var('compiler').lower()
        ldinfo = known_formats['native', 'dynamic']
        arinfo = known_formats['native', 'static']

        # Try to infer the appropriate -fuse-ld option from the LD environment
        # variable.
        link_command = command[:]
        ld_command = env.getvar(ldinfo.var('linker'))
        if ld_command:
            tail = os.path.splitext(ld_command)[1][1:]
            if tail in ['bfd', 'gold']:
                log.info('setting `-fuse-ld={}` for `{}`'
                         .format(tail, shell.join(command)))
                link_command.append('-fuse-ld={}'.format(tail))

        cflags_name = langinfo.var('flags').lower()
        cflags = (target_flags +
                  shell.split(env.getvar('CPPFLAGS', '')) +
                  shell.split(env.getvar(langinfo.var('flags'), '')))

        ldflags_name = ldinfo.var('flags').lower()
        ldflags = (target_flags +
                   shell.split(env.getvar(ldinfo.var('flags'), '')))
        ldlibs_name = ldinfo.var('libs').lower()
        ldlibs = shell.split(env.getvar(ldinfo.var('libs'), ''))

        ar_name = arinfo.var('linker').lower()
        ar_command = check_which(env.getvar(arinfo.var('linker'), 'ar'),
                                 env.variables, kind='static linker')
        arflags_name = arinfo.var('flags').lower()
        arflags = shell.split(env.getvar(arinfo.var('flags'), 'cr'))

        # macOS's ld doesn't support --version, but we can still try it out and
        # grab the command line.
        ld_command = None
        try:
            stdout, stderr = env.execute(
                command + ldflags + ['-v', '-Wl,--version'],
                stdout=shell.Mode.pipe, stderr=shell.Mode.pipe,
                returncode='any'
            )

            for line in stderr.split('\n'):
                if '--version' in line:
                    ld_command = shell.split(line)[0:1]
                    if os.path.basename(ld_command[0]) != 'collect2':
                        break
        except (OSError, shell.CalledProcessError):
            pass

        self.compiler = CcCompiler(self, env, name, command, cflags_name,
                                   cflags)
        try:
            self.pch_compiler = CcPchCompiler(self, env, name, command,
                                              cflags_name, cflags)
        except ValueError:
            self.pch_compiler = None

        self._linkers = {
            'executable': CcExecutableLinker(
                self, env, name, link_command, ldflags_name, ldflags,
                ldlibs_name, ldlibs
            ),
            'shared_library': CcSharedLibraryLinker(
                self, env, name, link_command, ldflags_name, ldflags,
                ldlibs_name, ldlibs
            ),
            'static_library': ArLinker(self, env, ar_name, ar_command,
                                       arflags_name, arflags),
        }
        if ld_command:
            self._linkers['raw'] = LdLinker(self, env, ld_command, stdout)

        self.packages = CcPackageResolver(self, env, command, ldflags)
        self.runner = None

    @classmethod
    def _parse_brand(cls, env, command, version_output):
        target_flags = []
        if 'Free Software Foundation' in version_output:
            brand = 'gcc'
            version = detect_version(version_output)
            if env.is_cross:
                triplet = parse_triplet(env.execute(
                    command + ['-dumpmachine'],
                    stdout=shell.Mode.pipe, stderr=shell.Mode.devnull
                ).rstrip())
                target_flags = cls._gcc_arch_flags(
                    env.target_platform.arch, triplet.arch
                )
        elif 'clang' in version_output:
            brand = 'clang'
            version = detect_version(version_output)
            if env.is_cross:
                target_flags = ['-target', env.target_platform.triplet]
        else:
            brand = 'unknown'
            version = None

        return brand, version, target_flags

    @staticmethod
    def _gcc_arch_flags(arch, native_arch):
        if arch == native_arch:
            return []
        elif arch == 'x86_64':
            return ['-m64']
        elif re.match(r'i.86$', arch):
            return ['-m32'] if not re.match(r'i.86$', native_arch) else []
        return []

    @staticmethod
    def check_command(env, command):
        return env.execute(command + ['--version'], stdout=shell.Mode.pipe,
                           stderr=shell.Mode.devnull)

    @property
    def flavor(self):
        return 'cc'

    @property
    def family(self):
        return 'native'

    @property
    def auto_link(self):
        return False

    @property
    def can_dual_link(self):
        return True

    def linker(self, mode):
        return self._linkers[mode]


class CcBaseCompiler(BuildCommand):
    def __init__(self, builder, env, rule_name, command_var, command,
                 cflags_name, cflags):
        BuildCommand.__init__(self, builder, env, rule_name, command_var,
                              command, flags=(cflags_name, cflags))

    @property
    def deps_flavor(self):
        return None if self.lang in ('f77', 'f95') else 'gcc'

    @property
    def needs_libs(self):
        return False

    def search_dirs(self, strict=False):
        return [abspath(i) for i in
                self.env.getvar('CPATH', '').split(os.pathsep)]

    def _call(self, cmd, input, output, deps=None, flags=None):
        result = list(chain(
            cmd, self._always_flags, iterate(flags), ['-c', input]
        ))
        if deps:
            result.extend(['-MMD', '-MF', deps])
        result.extend(['-o', output])
        return result

    @property
    def _always_flags(self):
        flags = ['-x', self._langs[self.lang]]
        # Force color diagnostics on Ninja, since it's off by default. See
        # <https://github.com/ninja-build/ninja/issues/174> for more
        # information.
        if self.env.backend == 'ninja':
            if self.brand == 'clang':
                flags.append('-fcolor-diagnostics')
            elif (self.brand == 'gcc' and self.version and
                  self.version in SpecifierSet('>=4.9')):
                flags.append('-fdiagnostics-color')
        return flags

    def _include_dir(self, directory):
        is_default = directory.path in self.env.host_platform.include_dirs

        # Don't include default directories as system dirs (e.g. /usr/include).
        # Doing so would break GCC 6 when #including stdlib.h:
        # <https://gcc.gnu.org/bugzilla/show_bug.cgi?id=70129>.
        if directory.system and not is_default:
            return ['-isystem', directory.path]
        else:
            return ['-I' + directory.path]

    def flags(self, options, output=None, mode='normal'):
        flags = []
        for i in options:
            if isinstance(i, opts.include_dir):
                flags.extend(self._include_dir(i.directory))
            elif isinstance(i, opts.define):
                if i.value:
                    flags.append('-D' + i.name + '=' + i.value)
                else:
                    flags.append('-D' + i.name)
            elif isinstance(i, opts.std):
                flags.append('-std=' + i.value)
            elif isinstance(i, opts.warning):
                for j in i.value:
                    if j == opts.WarningValue.disable:
                        flags.append('-w')
                    else:
                        flags.append('-W' + j.name)
            elif isinstance(i, opts.debug):
                flags.append('-g')
            elif isinstance(i, opts.optimize):
                for j in i.value:
                    flags.append(_optimize_flags[j])
            elif isinstance(i, opts.pthread):
                flags.append('-pthread')
            elif isinstance(i, opts.pic):
                flags.append('-fPIC')
            elif isinstance(i, opts.pch):
                flags.extend(['-include', i.header.path.stripext()])
            elif isinstance(i, opts.sanitize):
                flags.append('-fsanitize=address')
            elif isinstance(i, safe_str.stringy_types):
                flags.append(i)
            else:
                raise TypeError('unknown option type {!r}'.format(type(i)))
        return flags


class CcCompiler(CcBaseCompiler):
    _langs = {
        'c'     : 'c',
        'c++'   : 'c++',
        'objc'  : 'objective-c',
        'objc++': 'objective-c++',
        'f77'   : 'f77',
        'f95'   : 'f95',
        'java'  : 'java',
    }

    def __init__(self, builder, env, name, command, cflags_name, cflags):
        CcBaseCompiler.__init__(self, builder, env, name, name, command,
                                cflags_name, cflags)

    @property
    def accepts_pch(self):
        return True

    def default_name(self, input, context):
        return input.path.stripext().suffix

    def output_file(self, name, context):
        # XXX: MinGW's object format doesn't appear to be COFF...
        return ObjectFile(Path(name + '.o'), self.builder.object_format,
                          self.lang)


class CcPchCompiler(CcCompiler):
    _langs = {
        'c'     : 'c-header',
        'c++'   : 'c++-header',
        'objc'  : 'objective-c-header',
        'objc++': 'objective-c++-header',
    }

    def __init__(self, builder, env, name, command, cflags_name, cflags):
        if builder.lang not in self._langs:
            raise ValueError('{} has no precompiled headers'
                             .format(builder.lang))
        CcBaseCompiler.__init__(self, builder, env, name + '_pch', name,
                                command, cflags_name, cflags)

    @property
    def accepts_pch(self):
        # You can't pass a PCH to a PCH compiler!
        return False

    def default_name(self, input, context):
        return input.path.suffix

    def output_file(self, name, context):
        ext = '.gch' if self.brand == 'gcc' else '.pch'
        return PrecompiledHeader(Path(name + ext), self.lang)


class CcLinker(BuildCommand):
    __allowed_langs = {
        'c'     : {'c'},
        'c++'   : {'c', 'c++', 'f77', 'f95'},
        'objc'  : {'c', 'objc', 'f77', 'f95'},
        'objc++': {'c', 'c++', 'objc', 'objc++', 'f77', 'f95'},
        'f77'   : {'c', 'f77', 'f95'},
        'f95'   : {'c', 'f77', 'f95'},
        'java'  : {'java', 'c', 'c++', 'objc', 'objc++', 'f77', 'f95'},
    }

    def __init__(self, builder, env, rule_name, command_var, command,
                 ldflags_name, ldflags, ldlibs_name, ldlibs):
        BuildCommand.__init__(
            self, builder, env, rule_name, command_var, command,
            flags=(ldflags_name, ldflags), libs=(ldlibs_name, ldlibs)
        )

        # Create a regular expression to extract the library name for linking
        # with -l.
        lib_formats = [r'lib(.*)\.a']
        if not self.env.target_platform.has_import_library:
            so_ext = re.escape(self.env.target_platform.shared_library_ext)
            lib_formats.append(r'lib(.*)' + so_ext)
        self._lib_re = re.compile('(?:' + '|'.join(lib_formats) + ')$')

    def _extract_lib_name(self, library):
        basename = library.path.basename()
        m = self._lib_re.match(basename)
        if not m:
            raise ValueError("'{}' is not a valid library name"
                             .format(basename))

        # Get the first non-None group from the match.
        return next(i for i in m.groups() if i is not None)

    def can_link(self, format, langs):
        return (format == self.builder.object_format and
                self.__allowed_langs[self.lang].issuperset(langs))

    @property
    def needs_libs(self):
        return True

    @property
    def has_link_macros(self):
        # We only need to define LIBFOO_EXPORTS/LIBFOO_STATIC macros on
        # platforms that have different import/export rules for libraries. We
        # approximate this by checking if the platform uses import libraries,
        # and only define the macros if it does.
        return self.env.target_platform.has_import_library

    def sysroot(self, strict=False):
        try:
            # XXX: clang doesn't support -print-sysroot.
            return self.env.execute(
                self.command + self.global_flags + ['-print-sysroot'],
                stdout=shell.Mode.pipe, stderr=shell.Mode.devnull
            ).rstrip()
        except (OSError, shell.CalledProcessError):
            if strict:
                raise
            return '' if self.env.target_platform.family == 'windows' else '/'

    def search_dirs(self, strict=False):
        try:
            output = self.env.execute(
                self.command + self.global_flags + ['-print-search-dirs'],
                stdout=shell.Mode.pipe, stderr=shell.Mode.devnull
            )
            m = re.search(r'^libraries: =(.*)', output, re.MULTILINE)
            search_dirs = re.split(os.pathsep, m.group(1))

            # clang doesn't respect LIBRARY_PATH with -print-search-dirs;
            # see <https://bugs.llvm.org//show_bug.cgi?id=23877>.
            if self.brand == 'clang':
                search_dirs = (self.env.getvar('LIBRARY_PATH', '')
                               .split(os.pathsep)) + search_dirs
        except (OSError, shell.CalledProcessError):
            if strict:
                raise
            search_dirs = self.env.getvar('LIBRARY_PATH', '').split(os.pathsep)
        return [abspath(i) for i in search_dirs]

    def _call(self, cmd, input, output, libs=None, flags=None):
        return list(chain(
            cmd, self._always_flags, iterate(flags), iterate(input),
            iterate(libs), ['-o', output]
        ))

    def pre_build(self, build, name, context):
        entry_point = getattr(context, 'entry_point', None)
        return opts.option_list(opts.entry_point(entry_point) if entry_point
                                else None)

    @property
    def _always_flags(self):
        if self.builder.object_format == 'mach-o':
            return ['-Wl,-headerpad_max_install_names']
        return []

    def _local_rpath(self, library, output):
        if not isinstance(library, Library):
            return [], []

        runtime_lib = library.runtime_file
        if runtime_lib and self.builder.object_format == 'elf':
            path = runtime_lib.path.parent().cross(self.env)
            if path.root != Root.absolute and path.root not in InstallRoot:
                if not output:
                    raise ValueError('unable to construct rpath')
                path = path.relpath(output.path.parent(), prefix='$ORIGIN')
            rpath = [path]

            # Prior to binutils 2.28, GNU's BFD-based ld doesn't correctly
            # respect $ORIGIN in a shared library's DT_RPATH/DT_RUNPATH field.
            # This results in ld being unable to find other shared libraries
            # needed by the directly-linked library. For more information, see:
            # <https://sourceware.org/bugzilla/show_bug.cgi?id=20535>.
            try:
                ld = self.builder.linker('raw')
                fix_rpath = (ld.brand == 'bfd' and ld.version in
                             SpecifierSet('<2.28'))
            except KeyError:
                fix_rpath = False

            rpath_link = []
            if output and fix_rpath:
                rpath_link = [i.path.parent() for i in
                              recursive_walk(runtime_lib, 'runtime_deps')]

            return rpath, rpath_link

        # Either we don't need rpaths or the object format must not support
        # them, so just return nothing.
        return [], []

    def _installed_rpaths(self, options, output):
        result = []
        changed = False
        for i in options:
            if isinstance(i, opts.lib):
                lib = i.library
                if isinstance(lib, Library) and lib.runtime_file:
                    local = self._local_rpath(lib, output)[0][0]
                    installed = file_install_path(lib, cross=self.env).parent()
                    result.append(installed)
                    if not isinstance(local, BasePath) or local != installed:
                        changed = True
            elif isinstance(i, opts.rpath_dir):
                result.append(i.path)

        return uniques(result) if changed else []

    def _darwin_rpath(self, options, output):
        if output and self.builder.object_format == 'mach-o':
            # Currently, we set the rpath on macOS to make it easy to load
            # locally-built shared libraries. Once we install the build, we'll
            # convert all the rpath-based paths to absolute paths and remove
            # the rpath from the binary.
            for i in options:
                if ( isinstance(i, opts.lib) and
                     isinstance(i.library, Library) and not
                     isinstance(i.library, StaticLibrary) ):
                    return Path('.').cross(self.env).relpath(
                        output.path.parent(), prefix='@loader_path'
                    )

        # We didn't find a shared library, or we're just not be building for
        # macOS, so return nothing.
        return None

    def always_libs(self, primary):
        # XXX: Don't just asssume that these are the right libraries to use.
        # For instance, clang users might want to use libc++ instead.
        libs = opts.option_list()
        if self.lang in ('c++', 'objc++') and not primary:
            libs.append(opts.lib('stdc++'))
        if self.lang in ('objc', 'objc++'):
            libs.append(opts.lib('objc'))
        if self.lang in ('f77', 'f95') and not primary:
            libs.append(opts.lib('gfortran'))
        if self.lang == 'java' and not primary:
            libs.append(opts.lib('gcj'))
        return libs

    def _link_lib(self, library, raw_link):
        def common_link(library):
            return ['-l' + self._extract_lib_name(library)]

        if isinstance(library, WholeArchive):
            if self.env.target_platform.genus == 'darwin':
                return ['-Wl,-force_load', library.path]
            return ['-Wl,--whole-archive', library.path,
                    '-Wl,--no-whole-archive']
        elif isinstance(library, Framework):
            if not self.env.target_platform.has_frameworks:
                raise TypeError('frameworks not supported on this platform')
            return ['-framework', library.full_name]
        elif isinstance(library, string_types):
            return ['-l' + library]
        elif isinstance(library, SharedLibrary):
            # If we created this library, we know its soname is set, so passing
            # the raw path to the library works (without soname, the linker
            # would create a reference to the absolute path of the library,
            # which we don't want). We do this to avoid adding more `-L`
            # options than we really need, which makes it easier to find the
            # right library when there are name collisions (e.g. linking to a
            # system `libfoo` when also building a local `libfoo` to use
            # elsewhere).
            if raw_link and library.creator:
                return [library.path]
            return common_link(library)
        elif isinstance(library, StaticLibrary):
            # In addition to the reasons above for shared libraries, we pass
            # static libraries in raw form as a way of avoiding getting the
            # shared version when we don't want it. (There are linker options
            # that do this too, but this way is more compatible and fits with
            # what we already do.)
            if raw_link:
                return [library.path]
            return common_link(library)

        # If we get here, we should have a generic `Library` object (probably
        # from MinGW). The naming for these doesn't work with `-l`, but we'll
        # try just in case and back to emitting the path in raw form.
        try:
            return common_link(library)
        except ValueError:
            if raw_link:
                return [library.path]
            raise

    def _lib_dir(self, library, raw_link):
        if not isinstance(library, Library):
            return []
        elif isinstance(library, StaticLibrary):
            return [] if raw_link else [library.path.parent()]
        elif isinstance(library, SharedLibrary):
            return ([] if raw_link and library.creator else
                    [library.path.parent()])

        # As above, if we get here, we should have a generic `Library` object
        # (probably from MinGW). Use `-L` if the library name works with `-l`;
        # otherwise, return nothing, since the library itself will be passed
        # "raw" (like static libraries).
        try:
            self._extract_lib_name(library)
            return [library.path.parent()]
        except ValueError:
            if raw_link:
                return []
            raise

    def flags(self, options, output=None, mode='normal'):
        pkgconf_mode = mode == 'pkg-config'
        flags, rpaths, rpath_links, lib_dirs = [], [], [], []
        rpaths.extend(iterate(self._darwin_rpath(options, output)))

        for i in options:
            if isinstance(i, opts.lib_dir):
                lib_dirs.append(i.directory.path)
            elif isinstance(i, opts.lib):
                lib_dirs.extend(self._lib_dir(i.library, not pkgconf_mode))
                rp, rplink = self._local_rpath(i.library, output)
                rpaths.extend(rp)
                rpath_links.extend(rplink)
            elif isinstance(i, opts.rpath_dir):
                rpaths.append(i.path)
            elif isinstance(i, opts.rpath_link_dir):
                rpath_links.append(i.path)
            elif isinstance(i, opts.module_def):
                if self.env.target_platform.has_import_library:
                    flags.append(i.value.path)
            elif isinstance(i, opts.debug):
                flags.append('-g')
            elif isinstance(i, opts.optimize):
                for j in i.value:
                    flags.append(_optimize_flags[j])
            elif isinstance(i, opts.pthread):
                # macOS doesn't expect -pthread when linking.
                if self.env.target_platform.genus != 'darwin':
                    flags.append('-pthread')
            elif isinstance(i, opts.entry_point):
                if self.lang != 'java':
                    raise ValueError('entry point only applies to java')
                flags.append('--main={}'.format(i.value))
            elif isinstance(i, safe_str.stringy_types):
                flags.append(i)
            elif isinstance(i, opts.lib_literal):
                pass
            else:
                raise TypeError('unknown option type {!r}'.format(type(i)))

        flags.extend('-L' + i for i in uniques(lib_dirs))
        if not pkgconf_mode and rpaths:
            flags.append('-Wl,-rpath,' + safe_str.join(rpaths, ':'))
        if not pkgconf_mode and rpath_links:
            flags.append('-Wl,-rpath-link,' + safe_str.join(rpath_links, ':'))
        return flags

    def lib_flags(self, options, mode='normal'):
        pkgconf_mode = mode == 'pkg-config'
        flags = []
        for i in options:
            if isinstance(i, opts.lib):
                flags.extend(self._link_lib(i.library, not pkgconf_mode))
            elif isinstance(i, opts.lib_literal):
                flags.append(i.value)
        return flags

    def post_install(self, options, output, context):
        if self.builder.object_format not in ['elf', 'mach-o']:
            return None

        path = file_install_path(output)

        if self.builder.object_format == 'elf':
            rpath = self._installed_rpaths(options, output)
            return self.env.tool('patchelf')(path, rpath)
        else:  # mach-o
            rpath = self._darwin_rpath(options, output)
            changes = [(darwin_install_name(i),
                        file_install_path(i, cross=self.env))
                       for i in output.runtime_deps]
            return self.env.tool('install_name_tool')(
                path, path if self._is_library else None, rpath, changes
            )


class CcExecutableLinker(CcLinker):
    _is_library = False

    def __init__(self, builder, env, name, command, ldflags_name, ldflags,
                 ldlibs_name, ldlibs):
        CcLinker.__init__(self, builder, env, name + '_link', name, command,
                          ldflags_name, ldflags, ldlibs_name, ldlibs)

    def output_file(self, name, context):
        path = Path(name + self.env.target_platform.executable_ext)
        return Executable(path, self.builder.object_format, self.lang)


class CcSharedLibraryLinker(CcLinker):
    _is_library = True

    def __init__(self, builder, env, name, command, ldflags_name, ldflags,
                 ldlibs_name, ldlibs):
        CcLinker.__init__(self, builder, env, name + '_linklib', name, command,
                          ldflags_name, ldflags, ldlibs_name, ldlibs)

    @property
    def num_outputs(self):
        return 2 if self.env.target_platform.has_import_library else 'all'

    def _call(self, cmd, input, output, libs=None, flags=None):
        output = listify(output)
        result = CcLinker._call(self, cmd, input, output[0], libs, flags)
        if self.env.target_platform.has_import_library:
            result.append('-Wl,--out-implib=' + output[1])
        return result

    def _lib_name(self, name, prefix='lib', suffix=''):
        head, tail = Path(name).splitleaf()
        ext = self.env.target_platform.shared_library_ext
        return head.append(prefix + tail + ext + suffix)

    def post_build(self, build, options, output, context):
        if isinstance(output, VersionedSharedLibrary):
            # Make symlinks for the various versions of the shared lib.
            CopyFile(build, self.env, output.soname, output, mode='symlink')
            CopyFile(build, self.env, output.link, output.soname,
                     mode='symlink')
            return output.link

    def output_file(self, name, context):
        version = getattr(context, 'version', None)
        soversion = getattr(context, 'soversion', None)
        fmt = self.builder.object_format

        if version and self.env.target_platform.has_versioned_library:
            if self.env.target_platform.genus == 'darwin':
                real = self._lib_name(name + '.{}'.format(version))
                soname = self._lib_name(name + '.{}'.format(soversion))
            else:
                real = self._lib_name(name, suffix='.{}'.format(version))
                soname = self._lib_name(name, suffix='.{}'.format(soversion))
            link = self._lib_name(name)
            return VersionedSharedLibrary(real, fmt, self.lang, soname, link)
        elif self.env.target_platform.has_import_library:
            dllprefix = ('cyg' if self.env.target_platform.genus == 'cygwin'
                         else '')
            dllname = self._lib_name(name, prefix=dllprefix)
            impname = self._lib_name(name, suffix='.a')
            dll = DllBinary(dllname, fmt, self.lang, impname)
            return [dll, dll.import_lib]
        else:
            return SharedLibrary(self._lib_name(name), fmt, self.lang)

    @property
    def _always_flags(self):
        shared = ('-dynamiclib' if self.env.target_platform.genus == 'darwin'
                  else '-shared')
        return CcLinker._always_flags.fget(self) + [shared, '-fPIC']

    def _soname(self, library):
        if isinstance(library, VersionedSharedLibrary):
            soname = library.soname
        else:
            soname = library

        if self.env.target_platform.genus == 'darwin':
            return ['-install_name', darwin_install_name(soname)]
        else:
            return ['-Wl,-soname,' + soname.path.basename()]

    def compile_options(self, context):
        options = opts.option_list()
        if self.builder.object_format != 'coff':
            options.append(opts.pic())
        if self.has_link_macros:
            options.append(opts.define(library_macro(
                context.name, 'shared_library'
            )))
        return options

    def flags(self, options, output=None, mode='normal'):
        flags = CcLinker.flags(self, options, output, mode)
        if output:
            flags.extend(self._soname(first(output)))
        return flags


class CcPackageResolver(object):
    def __init__(self, builder, env, command, ldflags):
        self.builder = builder
        self.env = env

        self.include_dirs = [i for i in uniques(chain(
            self.builder.compiler.search_dirs(),
            self.env.host_platform.include_dirs
        )) if exists(i)]

        cc_lib_dirs = self.builder.linker('executable').search_dirs()
        try:
            sysroot = self.builder.linker('executable').sysroot()
            ld_lib_dirs = self.builder.linker('raw').search_dirs(sysroot, True)
        except (KeyError, OSError, shell.CalledProcessError):
            ld_lib_dirs = self.env.host_platform.lib_dirs

        self.lib_dirs = [i for i in uniques(chain(
            cc_lib_dirs, ld_lib_dirs, self.env.host_platform.lib_dirs
        )) if exists(i)]

    @property
    def lang(self):
        return self.builder.lang

    def header(self, name, search_dirs=None):
        if search_dirs is None:
            search_dirs = self.include_dirs

        for base in search_dirs:
            if base.root != Root.absolute:
                raise ValueError('expected an absolute path')
            if exists(base.append(name)):
                return HeaderDirectory(base, None, system=True)

        raise PackageResolutionError("unable to find header '{}'".format(name))

    def library(self, name, kind=PackageKind.any, search_dirs=None):
        if search_dirs is None:
            search_dirs = self.lib_dirs

        libnames = []
        if kind & PackageKind.shared:
            base = 'lib' + name + self.env.target_platform.shared_library_ext
            if self.env.target_platform.has_import_library:
                libnames.append((base + '.a', LinkLibrary, {}))
            else:
                libnames.append((base, SharedLibrary, {}))
        if kind & PackageKind.static:
            libnames.append(('lib' + name + '.a', StaticLibrary,
                             {'lang': self.lang}))

        # XXX: Include Cygwin here too?
        if self.env.target_platform.family == 'windows':
            # We don't actually know what kind of library this is. It could be
            # a static library or an import library (which we classify as a
            # kind of shared lib).
            libnames.append((name + '.lib', Library, {}))

        for base in search_dirs:
            if base.root != Root.absolute:
                raise ValueError('expected an absolute path')
            for libname, libkind, extra_kwargs in libnames:
                fullpath = base.append(libname)
                if exists(fullpath):
                    return libkind(fullpath, format=self.builder.object_format,
                                   **extra_kwargs)

        raise PackageResolutionError("unable to find library '{}'"
                                     .format(name))

    def resolve(self, name, version, kind, headers, lib_names):
        format = self.builder.object_format
        try:
            return pkg_config.resolve(self.env, name, format, version, kind)
        except (OSError, PackageResolutionError):
            compile_options = opts.option_list()
            link_options = opts.option_list()

            compile_options.extend(opts.include_dir(self.header(i))
                                   for i in iterate(headers))

            if lib_names is default_sentinel:
                lib_names = self.env.target_platform.transform_package(name)
            for i in iterate(lib_names):
                if isinstance(i, Framework):
                    link_options.append(opts.lib(i))
                elif i == 'pthread':
                    compile_options.append(opts.pthread())
                    link_options.append(opts.pthread())
                else:
                    link_options.append(opts.lib(self.library(i, kind)))

            return CommonPackage(name, format, compile_options, link_options)



import sys

from enum import Enum

from .app_version import version
from .arguments import parser as argparse

# Munge the depfile so that it works a little better under Make. Specifically,
# we need all the dependencies in the depfile to also be targets, so that we
# don't get an error if a dep is removed. For a more-detailed discussion of why
# this is necessary, see <http://scottmcpeak.com/autodepend/autodepend.html>.

Token = Enum('Token', ['char', 'colon', 'space', 'newline'])
State = Enum('State', ['target', 'between_targets', 'dep', 'between_deps'])


class ParseError(ValueError):
    pass


class UnexpectedTokenError(ParseError):
    def __init__(self, tok):
        ParseError.__init__(self, "unexpected token '{}'".format(tok))


def tokenize(s):
    # The depfile syntax is a bit weird, since it seems no one quite
    # understands the correct ways to escape characters for Make in all cases
    # (made worse by the fact that even GNU Make's behavior varies across
    # versions). For our purposes though, we only need to recognize when
    # unescaped colons (always followed by whitespace in the depfile
    # generators) and unescaped spaces are emitted.

    s = iter(s)
    while True:
        c = next(s, None)
        if c is None:
            return

        if c == ':':
            c = next(s, None)
            if c is None or c in ' \t\n':
                yield (Token.colon, None)
                if c is None:
                    return
            else:
                yield (Token.char, ':')

        if c == '\\':
            c = next(s, None)
            if c != '\n':  # Swallow escaped newlines.
                yield (Token.char, '\\')
                if c is None:
                    return
                yield (Token.char, c)
        elif c in ' \t':
            yield (Token.space, None)
        elif c == '\n':
            yield (Token.newline, None)
        else:
            yield (Token.char, c)


def emit_deps(instream, outstream):
    state = State.target

    for tok, value in tokenize(instream.read()):
        if state == State.target:
            if tok == Token.space:
                state = State.between_targets
            elif tok == Token.colon:
                state = State.between_deps
            elif tok != Token.char:
                raise UnexpectedTokenError(tok)
        elif state == State.between_targets:
            if tok == Token.char:
                state = State.target
            elif tok == Token.colon:
                state = State.between_deps
            elif tok != Token.space:
                raise UnexpectedTokenError(tok)
        elif state == State.dep:
            if tok == Token.char:
                outstream.write(value)
            elif tok == Token.space:
                outstream.write(':\n')
                state = State.between_deps
            elif tok == Token.newline:
                outstream.write(':\n')
                state = State.target
            else:
                raise UnexpectedTokenError(tok)
        else:  # state == State.between_deps
            if tok == Token.char:
                state = State.dep
                outstream.write(value)
            elif tok == Token.newline:
                state = State.target
            elif tok != Token.space:
                raise UnexpectedTokenError(tok)

    if state != State.target:
        raise ParseError('unexpected end of file')


def main():
    parser = argparse.ArgumentParser(
        prog='bfg9000-depfixer',
        description='Read in a depfile (in Makefile syntax) on stdin and ' +
                    'output all the dependencies as targets on stdout.'
    )
    parser.add_argument('--version', action='version',
                        version='%(prog)s ' + version)
    parser.parse_args()

    try:
        emit_deps(sys.stdin, sys.stdout)
    except Exception as e:
        parser.error(e)



import json
import os
import re
import subprocess
from setuptools import setup, find_packages, Command

from bfg9000.app_version import version

root_dir = os.path.abspath(os.path.dirname(__file__))


class Coverage(Command):
    description = 'run tests with code coverage'
    user_options = [
        ('test-suite=', 's',
         "test suite to run (e.g. 'some_module.test_suite')"),
    ]

    def initialize_options(self):
        self.test_suite = None

    def finalize_options(self):
        pass

    def run(self):
        env = dict(os.environ)
        pythonpath = os.path.join(root_dir, 'test', 'scripts')
        if env.get('PYTHONPATH'):
            pythonpath += os.pathsep + env['PYTHONPATH']
        env.update({
            'PYTHONPATH': pythonpath,
            'COVERAGE_FILE': os.path.join(root_dir, '.coverage'),
            'COVERAGE_PROCESS_START': os.path.join(root_dir, '.coveragerc'),
        })

        subprocess.check_call(['coverage', 'erase'])
        subprocess.check_call(
            ['coverage', 'run', 'setup.py', 'test'] +
            (['-q'] if self.verbose == 0 else []) +
            (['-s', self.test_suite] if self.test_suite else []),
            env=env
        )
        subprocess.check_call(['coverage', 'combine'])


custom_cmds = {
    'coverage': Coverage,
}

try:
    from packaging.version import Version

    class DocServe(Command):
        description = 'serve the documentation locally'
        user_options = [
            ('working', 'w', 'use the documentation in the working directory'),
            ('dev-addr=', None, 'address to host the documentation on'),
        ]

        def initialize_options(self):
            self.working = False
            self.dev_addr = '0.0.0.0:8000'

        def finalize_options(self):
            pass

        def run(self):
            cmd = 'mkdocs' if self.working else 'mike'
            subprocess.check_call([
                cmd, 'serve', '--dev-addr=' + self.dev_addr
            ])

    class DocDeploy(Command):
        description = 'push the documentation to GitHub'
        user_options = []

        def initialize_options(self):
            pass

        def finalize_options(self):
            pass

        def run(self):
            v = Version(version)
            alias = 'dev' if v.is_devrelease else 'latest'
            title = '{} ({})'.format(v.base_version, alias)
            short_version = '{}.{}'.format(*v.release[:2])

            info = json.loads(subprocess.check_output(
                ['mike', 'list', '-j', alias],
                universal_newlines=True
            ))
            if info['version'] != short_version:
                t = re.sub(r' \({}\)$'.format(re.escape(alias)), '',
                           info['title'])
                subprocess.check_call(['mike', 'retitle', info['version'], t])

            subprocess.check_call(['mike', 'deploy', '-ut', title,
                                   short_version, alias])

    custom_cmds['doc_serve'] = DocServe
    custom_cmds['doc_deploy'] = DocDeploy
except ImportError:
    pass

try:
    from flake8.main.setuptools_command import Flake8

    class LintCommand(Flake8):
        def distribution_files(self):
            return ['setup.py', 'bfg9000', 'examples', 'test']

    custom_cmds['lint'] = LintCommand
except ImportError:
    pass

more_requires = []

if os.getenv('STDEB_BUILD') not in ['1', 'true']:
    more_requires.extend([
        'enum34;python_version<"3.4"',
        'doppel >= 0.4.0',
        'pysetenv;platform_system=="Windows"'
    ])
    if os.getenv('NO_PATCHELF') not in ['1', 'true']:
        more_requires.append('patchelf-wrapper;platform_system=="Linux"')

with open(os.path.join(root_dir, 'README.md'), 'r') as f:
    # Read from the file and strip out the badges.
    long_desc = re.sub(r'(^# bfg9000.*)\n\n(.+\n)*', r'\1', f.read())

try:
    import pypandoc
    long_desc = pypandoc.convert(long_desc, 'rst', format='md')
except ImportError:
    pass

setup(
    name='bfg9000',
    version=version,

    description='A cross-platform build file generator',
    long_description=long_desc,
    keywords='build file generator',
    url='https://jimporter.github.io/bfg9000/',

    author='Jim Porter',
    author_email='itsjimporter@gmail.com',
    license='BSD',

    classifiers=[
        'Development Status :: 3 - Alpha',

        'Intended Audience :: Developers',

        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: BSD License',

        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],

    packages=find_packages(exclude=['test', 'test.*']),

    install_requires=(
        ['colorama', 'packaging >= 17.0', 'setuptools', 'six'] +
        more_requires
    ),
    extras_require={
        'dev': ['coverage', 'flake8 >= 3.0', 'lxml', 'mike >= 0.3.1',
                'mkdocs-bootswatch', 'mock', 'pypandoc', 'stdeb'],
        'test': ['coverage', 'flake8 >= 3.0', 'lxml', 'mock'],
        'msbuild': ['lxml'],
    },

    entry_points={
        'console_scripts': [
            'bfg9000=bfg9000.driver:main',
            '9k=bfg9000.driver:simple_main',
            'bfg9000-depfixer=bfg9000.depfixer:main',
            'bfg9000-jvmoutput=bfg9000.jvmoutput:main',
            'bfg9000-rccdep=bfg9000.rccdep:main',
        ],
        'bfg9000.backends': [
            'make=bfg9000.backends.make.writer',
            'ninja=bfg9000.backends.ninja.writer',
            'msbuild=bfg9000.backends.msbuild.writer [msbuild]',
        ],
        'bfg9000.platforms.host': [
            'cygwin=bfg9000.platforms.cygwin:CygwinHostPlatform',
            'darwin=bfg9000.platforms.posix:PosixHostPlatform',
            'linux=bfg9000.platforms.posix:PosixHostPlatform',
            'msdos=bfg9000.platforms.windows:WindowsTargetPlatform',
            'posix=bfg9000.platforms.posix:PosixHostPlatform',
            'win9x=bfg9000.platforms.windows:WindowsHostPlatform',
            'winnt=bfg9000.platforms.windows:WindowsHostPlatform',
        ],
        'bfg9000.platforms.target': [
            'cygwin=bfg9000.platforms.cygwin:CygwinTargetPlatform',
            'darwin=bfg9000.platforms.posix:DarwinTargetPlatform',
            'linux=bfg9000.platforms.posix:PosixTargetPlatform',
            'msdos=bfg9000.platforms.windows:WindowsTargetPlatform',
            'posix=bfg9000.platforms.posix:PosixTargetPlatform',
            'win9x=bfg9000.platforms.windows:WindowsTargetPlatform',
            'winnt=bfg9000.platforms.windows:WindowsTargetPlatform',
        ],
    },

    test_suite='test',
    cmdclass=custom_cmds,
)



from collections import OrderedDict
from pkg_resources import iter_entry_points, DistributionNotFound

from ..objutils import memoize


@memoize
def list_backends():
    backends = []
    for i in iter_entry_points('bfg9000.backends'):
        try:
            backend = i.load()
            backends.append((i.name, backend))
        # An ImportError can be thrown by the MSBuild backend when bfg9000 was
        # installed as a Wheel with setuptools < 34.0.2. We'd require 34.0.2 or
        # better, but setuptools < 11.0 won't let you upgrade setuptools during
        # bfg's installation. Since Ubuntu 14.04 is supported until 2019 and
        # has setuptools 2.2 by default, we're stuck with this for a while.
        except (DistributionNotFound, ImportError):  # pragma: no cover
            pass

    def sort_key(x):
        return x[1].priority if x[1].version() else 0
    backends.sort(key=sort_key, reverse=True)
    return OrderedDict(backends)



from .basepath import BasePath, Root, InstallRoot
from .core import Platform
from .framework import Framework
from .host import HostPlatform
from .target import TargetPlatform


class PosixPath(BasePath):
    def _localize_path(self, path):
        return path


class PosixPlatform(Platform):
    @property
    def _triplet_vendor(self):
        if self.genus == 'darwin':
            return 'apple'
        return Platform._triplet_vendor.fget(self)

    @property
    def _triplet_sys_abi(self):
        if self.genus == 'darwin':
            return 'darwin'
        elif self.genus == 'linux':
            abi = 'android' if self.species == 'android' else 'gnu'
            return 'linux-' + abi
        elif self.genus == 'cygwin':
            return 'windows-cygnus'
        return self.species

    @property
    def family(self):
        return 'posix'

    Path = PosixPath


class PosixHostPlatform(HostPlatform, PosixPlatform):
    @property
    def include_dirs(self):
        return [PosixPath('/usr/local/include', Root.absolute),
                PosixPath('/usr/include', Root.absolute)]

    @property
    def lib_dirs(self):
        return [PosixPath('/usr/local/lib', Root.absolute),
                PosixPath('/lib', Root.absolute),
                PosixPath('/usr/lib', Root.absolute)]

    @property
    def destdir(self):
        return True


class PosixTargetPlatform(TargetPlatform, PosixPlatform):
    _package_map = {
        'gl': 'GL',
        'glu': 'GLU',
        'zlib': 'z',
    }

    @property
    def object_format(self):
        return 'elf'

    @property
    def executable_ext(self):
        return ''

    @property
    def shared_library_ext(self):
        return '.so'

    @property
    def has_import_library(self):
        return False

    @property
    def has_versioned_library(self):
        return True

    @property
    def has_frameworks(self):
        return False

    @property
    def install_dirs(self):
        return {
            InstallRoot.prefix     : PosixPath('/usr/local', Root.absolute),
            InstallRoot.exec_prefix: PosixPath('', InstallRoot.prefix),
            InstallRoot.bindir     : PosixPath('bin', InstallRoot.exec_prefix),
            InstallRoot.libdir     : PosixPath('lib', InstallRoot.exec_prefix),
            InstallRoot.includedir : PosixPath('include', InstallRoot.prefix),
        }


class DarwinTargetPlatform(PosixTargetPlatform):
    _package_map = {
        'gl': Framework('OpenGL'),
        'glu': Framework('OpenGL'),
        'glut': Framework('GLUT'),
    }

    @property
    def object_format(self):
        return 'mach-o'

    @property
    def shared_library_ext(self):
        return '.dylib'

    @property
    def has_frameworks(self):
        return True



import errno
import functools
import os
from contextlib import contextmanager

from .platforms.basepath import BasePath, Root, InstallRoot, DestDir  # noqa
from .platforms.host import platform_info

Path = platform_info().Path


def abspath(path, type=Path):
    return type.abspath(path)


def install_path(path, install_root, directory=False, cross=None):
    cls = cross.target_platform.Path if cross else type(path)
    if path.root == Root.absolute:
        if not cross:
            raise TypeError('path is absolute')
        return cls(path.suffix, path.root)

    if path.root == Root.srcdir:
        suffix = path.curdir if directory else path.basename()
    else:
        suffix = path.suffix
    return cls(suffix, install_root, destdir=not cross)


def commonprefix(paths):
    if not paths or any(i.root != paths[0].root for i in paths):
        return None

    cls = type(paths[0])
    split = [i.split() for i in paths]
    lo, hi = min(split), max(split)

    for i, bit in enumerate(lo):
        if bit != hi[i]:
            return cls(cls.sep.join(lo[:i]), paths[0].root)
    return cls(cls.sep.join(lo), paths[0].root)


def _wrap_ospath(fn):
    @functools.wraps(fn)
    def wrapper(path, variables=None):
        return fn(path.string(variables))

    return wrapper


exists = _wrap_ospath(os.path.exists)
isdir = _wrap_ospath(os.path.isdir)
isfile = _wrap_ospath(os.path.isfile)
islink = _wrap_ospath(os.path.islink)


def samefile(path1, path2, variables=None):
    if hasattr(os.path, 'samefile'):
        return os.path.samefile(path1.string(variables),
                                path2.string(variables))
    else:
        # This isn't entirely accurate, but it's close enough, and should only
        # be necessary for Windows with Python 2.x.
        return (os.path.realpath(path1.string(variables)) ==
                os.path.realpath(path2.string(variables)))


def makedirs(path, mode=0o777, exist_ok=False):
    try:
        os.makedirs(path, mode)
    except OSError as e:
        if not exist_ok or e.errno != errno.EEXIST or not os.path.isdir(path):
            raise


# Make an alias since the function below masks the module-level function with
# one of its parameters.
_makedirs = makedirs


@contextmanager
def pushd(dirname, makedirs=False, mode=0o777, exist_ok=False):
    old = os.getcwd()
    if makedirs:
        _makedirs(dirname, mode, exist_ok)

    os.chdir(dirname)
    yield
    os.chdir(old)



import platform
import re
from itertools import chain
from packaging.specifiers import (
    LegacySpecifier as Specifier, Specifier as PythonSpecifier,
    SpecifierSet as PythonSpecifierSet
)
from packaging.version import (
    Version as PythonVersion, LegacyVersion as Version
)

from .app_version import version as bfg_version
from .exceptions import VersionError
from .iterutils import iterate

__all__ = ['bfg_version', 'check_version', 'detect_version', 'python_version',
           'PythonSpecifier', 'PythonSpecifierSet', 'PythonVersion',
           'simplify_specifiers', 'Specifier', 'SpecifierSet', 'Version',
           'VersionError']

bfg_version = PythonVersion(bfg_version)
# Strip trailing "+" from Python version. Some versions in distros have this...
python_version = PythonVersion(re.sub(r'\+$', '', platform.python_version()))


# Use a LegacySpecifierSet instead once packaging.specifiers has it. See
# <https://github.com/pypa/packaging/pull/92>.
class SpecifierSet(PythonSpecifierSet):
    def __init__(self, specifiers=''):
        specifiers = [s.strip() for s in specifiers.split(',') if s.strip()]
        parsed = set()
        for specifier in specifiers:
            parsed.add(Specifier(specifier))
        self._specs = frozenset(parsed)
        self._prereleases = None


def simplify_specifiers(spec):
    """Try to simplify a SpecifierSet by combining redundant specifiers."""

    def key(s):
        return (s.version, 1 if s.operator in ['>=', '<'] else 2)

    def in_bounds(v, lo, hi):
        if lo and v not in lo:
            return False
        if hi and v not in hi:
            return False
        return True

    def err(reason='inconsistent'):
        return ValueError('{} specifier set {}'.format(reason, spec))

    gt = None
    lt = None
    eq = None
    ne = []

    for i in spec:
        if i.operator == '==':
            if eq is None:
                eq = i
            elif eq != i:  # pragma: no branch
                raise err()
        elif i.operator == '!=':
            ne.append(i)
        elif i.operator in ['>', '>=']:
            gt = i if gt is None else max(gt, i, key=key)
        elif i.operator in ['<', '<=']:
            lt = i if lt is None else min(lt, i, key=key)
        else:
            raise err('invalid')

    ne = [i for i in ne if in_bounds(i.version, gt, lt)]
    if eq:
        if ( any(i.version in eq for i in ne) or
             not in_bounds(eq.version, gt, lt)):
            raise err()
        return SpecifierSet(str(eq))
    if lt and gt:
        if lt.version not in gt or gt.version not in lt:
            raise err()
        if ( gt.version == lt.version and gt.operator == '>=' and
             lt.operator == '<='):
            return SpecifierSet('=={}'.format(gt.version))

    return SpecifierSet(
        ','.join(str(i) for i in chain(iterate(gt), iterate(lt), ne))
    )


def check_version(version, specifier, kind, exception_type=VersionError):
    msg = "{kind} version {ver} doesn't meet requirement {req}"
    if version not in specifier:
        raise exception_type(msg.format(kind=kind, ver=version, req=specifier))


def detect_version(string, pre='', post='', flags=0):
    m = re.search(pre + r'(\d+(?:\.\d+)+)' + post, string, flags)
    return Version(m.group(1)) if m else None



from . import tool
from .common import SimpleCommand
from ..iterutils import flatten, listify


@tool('install_name_tool')
class InstallNameTool(SimpleCommand):
    rule_name = command_var = 'install_name_tool'

    def __init__(self, env):
        SimpleCommand.__init__(
            self, env, name='install_name_tool', env_var='INSTALL_NAME_TOOL',
            default='install_name_tool'
        )

    def _call(self, cmd, file, id=None, delete_rpath=None, changes=[]):
        args = []
        if id:
            args += ['-id', id]
        if delete_rpath:
            args += ['-delete_rpath', delete_rpath]
        args += flatten(['-change'] + listify(i) for i in changes)

        if args:
            return cmd + args + [file]



import os.path

from . import *


class TestOptions(IntegrationTest):
    def __init__(self, *args, **kwargs):
        IntegrationTest.__init__(
            self, os.path.join(examples_dir, '03_options'), *args, **kwargs
        )

    def test_build(self):
        self.build()
        self.assertOutput([executable('program')], 'hello, world!\n')
        if env.host_platform.genus == 'linux':
            output = self.assertPopen(['readelf', '-s', executable('program')])
            assertNotRegex(self, output, r"Symbol table '\.symtab'")



import ntpath
import os
from collections import defaultdict
from enum import Enum
from itertools import chain
from lxml import etree
from lxml.builder import E
from six import iteritems, string_types

from .solution import uuid_str
from ... import path
from ... import safe_str
from ...shell import windows as wshell
from ...file_types import File
from ...iterutils import isiterable
from ...tools.common import Command

__all__ = ['BuildDir', 'CommandProject', 'NoopProject', 'Project',
           'VcxProject', 'textify', 'textify_each']

BuildDir = Enum('BuildDir', ['output', 'intermediate', 'solution'])


_path_vars = {
    BuildDir.output: {
        path.Root.srcdir: '$(SourceDir)',
        path.Root.builddir: '$(OutDir)',
    },
    BuildDir.intermediate: {
        path.Root.srcdir: '$(SourceDir)',
        path.Root.builddir: '$(IntDir)',
    },
    BuildDir.solution: {
        path.Root.srcdir: '$(SourceDir)',
        path.Root.builddir: '$(SolutionDir)',
    },
}


def textify(thing, quoted=False, builddir=BuildDir.output):
    if isinstance(thing, File):
        if thing.creator is None:
            builddir = BuildDir.solution
        elif getattr(thing.creator, 'msbuild_output', False):
            builddir = BuildDir.output
        else:
            builddir = BuildDir.intermediate

    thing = safe_str.safe_str(thing)
    if isinstance(thing, safe_str.literal_types):
        return thing.string
    elif isinstance(thing, string_types):
        return wshell.quote(thing, escape_percent=True) if quoted else thing
    elif isinstance(thing, safe_str.jbos):
        return ''.join(textify(i, quoted, builddir) for i in thing.bits)
    elif isinstance(thing, path.BasePath):
        return ntpath.normpath(thing.realize(_path_vars[builddir]))
    else:
        raise TypeError(type(thing))


def textify_each(thing, *args, **kwargs):
    return (textify(i, *args, **kwargs) for i in thing)


class Project(object):
    _XMLNS = 'http://schemas.microsoft.com/developer/msbuild/2003'
    _DOCTYPE = '<?xml version="1.0" encoding="utf-8"?>'
    _extension = '.proj'

    def __init__(self, env, name, configuration=None, dependencies=None):
        self.name = name
        self.uuid = None
        self.configuration = configuration or 'Debug'
        self.dependencies = dependencies or []

        self.version = env.getvar('VISUALSTUDIOVERSION', '14.0')

        # Some projects map one platform name to another in the .sln file (e.g.
        # "x86" => "Win32"). I'm not sure why this is, but we replicate it via
        # a "platform" and "real platform" so that C++ builds work from a VS
        # 2017 command prompt.
        self.platform = env.getvar('PLATFORM', 'Win32')
        self.real_platform = self.platform
        self.srcdir = env.srcdir

    @property
    def path(self):
        return path.Path(self.name).append('{name}{ext}'.format(
            name=os.path.basename(self.name), ext=self._extension
        ))

    def set_uuid(self, uuids):
        if not self.uuid:  # pragma: no branch
            self.uuid = uuids[self.name]

    @property
    def uuid_str(self):
        return uuid_str(self.uuid)

    @property
    def config_plat(self):
        return '{config}|{platform}'.format(
            config=self.configuration,
            platform=self.platform
        )

    @property
    def real_config_plat(self):
        return '{config}|{platform}'.format(
            config=self.configuration,
            platform=self.real_platform
        )

    def _write(self, out, children):
        project = E.Project({'DefaultTargets': 'Build',
                             'ToolsVersion': self.version,
                             'xmlns': self._XMLNS},
            E.ItemGroup({'Label': 'ProjectConfigurations'},
                E.ProjectConfiguration({'Include' : self.real_config_plat},
                    E.Configuration(self.configuration),
                    E.Platform(self.real_platform)
                )
            ),
            E.PropertyGroup({'Label': 'Globals'},
                E.ProjectGuid(self.uuid_str),
                E.RootNamespace(self.name),
                E.Platform(self.real_platform),
                E.SourceDir(textify(self.srcdir))
            ),
            *children
        )
        out.write(etree.tostring(project, doctype=self._DOCTYPE,
                                 pretty_print=True))


class VcxProject(Project):
    _extension = '.vcxproj'

    _warning_levels = {
        '0'  : 'TurnOffAllWarnings',
        '1'  : 'Level1',
        '2'  : 'Level2',
        '3'  : 'Level3',
        '4'  : 'Level4',
        'all': 'EnableAllWarnings',
    }

    def __init__(self, env, name, mode='Application', configuration=None,
                 output_file=None, files=None, compile_options=None,
                 link_options=None, dependencies=None):
        Project.__init__(self, env, name, configuration, dependencies)
        self.mode = mode
        self.output_file = output_file
        self.files = files or []
        self.compile_options = compile_options or {}
        self.link_options = link_options or {}

        version = env.getvar('VCTOOLSVERSION', self.version)
        self.toolset = 'v' + version.replace('.', '')[0:3]

        # As above, VS 2017 remaps x86 to Win32 for C++ projects.
        if self.real_platform == 'x86':
            self.real_platform = 'Win32'

    def write(self, out):
        target_name = safe_str.safe_str(self.output_file).basename()
        override_props = E.PropertyGroup(
            E.TargetName(os.path.splitext(target_name)[0]),
            E.TargetPath(textify(self.output_file))
        )

        compile_opts = E.ClCompile()
        self._write_compile_options(compile_opts, self.compile_options)
        link_opts = E.Lib() if self.mode == 'StaticLibrary' else E.Link()
        self._write_link_options(link_opts, self.link_options)

        self._write(out, [
            E.Import(Project=r'$(VCTargetsPath)\Microsoft.Cpp.default.props'),
            E.PropertyGroup({'Label': 'Configuration'},
                E.ConfigurationType(self.mode),
                E.UseDebugLibraries('true'),
                E.PlatformToolset(self.toolset),
                E.CharacterSet('Multibyte')
            ),
            E.Import(Project=r'$(VCTargetsPath)\Microsoft.Cpp.props'),
            override_props,
            E.ItemDefinitionGroup(compile_opts, link_opts),
            self._compiles(self.files),
            E.Import(Project=r'$(VCTargetsPath)\Microsoft.Cpp.Targets')
        ])

    def _compiles(self, files):
        def basename(path):
            return path.stripext().basename()

        # First, figure out all the common prefixes of files with the same
        # basename so we can use this to give them unique object names.
        names = defaultdict(lambda: [])
        for i in self.files:
            p = i['name'].path
            names[basename(p)].append(p.parent())
        prefixes = {k: path.commonprefix(v) if len(v) > 1 else None
                    for k, v in iteritems(names)}

        compiles = E.ItemGroup()
        for i in self.files:
            name = i['name']
            c = E.ClCompile(Include=textify(name))
            self._write_compile_options(c, i['options'])

            prefix = prefixes[basename(name.path)]
            if prefix:
                # If this prefix is shared with another file, strip it out to
                # create a unique directory to store this object file.
                suffix = path.Path(name.path.relpath(prefix)).stripext('.obj')
                c.append(E.ObjectFileName(textify(
                    suffix, builddir=BuildDir.intermediate
                )))
            compiles.append(c)
        return compiles

    def _write_compile_options(self, element, options):
        warnings = options.get('warnings', {})
        if warnings.get('level') is not None:
            element.append(E.WarningLevel(
                self._warning_levels[warnings['level']]
            ))
        if warnings.get('as_error') is not None:
            element.append(E.TreatWarningAsError(
                'true' if warnings['as_error'] else 'false'
            ))

        if options.get('includes'):
            element.append(E.AdditionalIncludeDirectories( ';'.join(chain(
                textify_each(options['includes']),
                ['%(AdditionalIncludeDirectories)']
            )) ))

        if options.get('defines'):
            element.append(E.PreprocessorDefinitions( ';'.join(chain(
                textify_each(options['defines']),
                ['%(PreprocessorDefinitions)']
            )) ))

        pch = options.get('pch', {})
        if pch.get('create') is not None:
            element.append(E.PrecompiledHeader('Create'))
            element.append(E.PrecompiledHeaderFile(pch['create']))
        elif pch.get('use') is not None:
            element.append(E.PrecompiledHeader('Use'))
            element.append(E.PrecompiledHeaderFile(pch['use']))

        if options.get('extra'):
            element.append(E.AdditionalOptions( ' '.join(chain(
                textify_each(options['extra'], quoted=True),
                ['%(AdditionalOptions)']
            )) ))

    def _write_link_options(self, element, options):
        element.append(E.OutputFile('$(TargetPath)'))

        if options.get('import_lib'):
            element.append(E.ImportLibrary(
                textify(options['import_lib'])
            ))

        if options.get('extra'):
            element.append(E.AdditionalOptions( ' '.join(chain(
                textify_each(options['extra'], quoted=True),
                ['%(AdditionalOptions)']
            )) ))

        if options.get('libs'):
            element.append(E.AdditionalDependencies( ';'.join(chain(
                textify_each(options['libs']),
                ['%(AdditionalDependencies)']
            )) ))


class NoopProject(Project):
    def write(self, out):
        self._write(out, [E.Target(Name='Build')])


class CommandProject(Project):
    def __init__(self, env, name, configuration=None, commands=None,
                 dependencies=None, makedir='$(OutDir)'):
        Project.__init__(self, env, name, configuration, dependencies)
        self.commands = commands or []
        self.makedir = makedir

    @staticmethod
    def convert_attr(value):
        if isiterable(value):
            return ';'.join(textify_each(value, quoted=True))
        else:
            return textify(value)

    @staticmethod
    def convert_command(value):
        value = Command.convert_args(value, lambda x: x.command)
        return ' '.join(textify_each(value, quoted=True))

    @classmethod
    def task(cls, task, **kwargs):
        return E(task, **{k: cls.convert_attr(v)
                          for k, v in iteritems(kwargs)})

    def write(self, out):
        target = E.Target(Name='Build')
        if self.makedir:
            target.append(E.MakeDir(Directories=self.makedir))

        for line in self.commands:
            if not isinstance(line, etree._Element):  # pragma: no cover
                raise TypeError('expected an lxml element')
            target.append(line)

        self._write(out, [
            # Import the C++ properties to get $(OutDir). There might be a
            # better way to handle this.
            E.Import(Project=r'$(VCTargetsPath)\Microsoft.Cpp.default.props'),
            E.Import(Project=r'$(VCTargetsPath)\Microsoft.Cpp.props'),
            target
        ])



import os
from itertools import chain

from .. import options as opts, safe_str, shell
from .common import BuildCommand, library_macro
from ..file_types import StaticLibrary
from ..iterutils import iterate
from ..objutils import memoize
from ..path import Path
from ..versioning import detect_version


class ArLinker(BuildCommand):
    def __init__(self, builder, env, name, command, arflags_name, arflags):
        BuildCommand.__init__(self, builder, env, name, name, command,
                              flags=(arflags_name, arflags))

    @memoize
    def _check_version(self):
        try:
            output = self.env.execute(
                self.command + ['--version'], stdout=shell.Mode.pipe,
                stderr=shell.Mode.devnull
            )
            if 'GNU ar' in output:
                return 'gnu', detect_version(output)
        except (OSError, shell.CalledProcessError):
            pass
        return 'unknown', None

    @property
    def brand(self):
        return self._check_version()[0]

    @property
    def version(self):
        return self._check_version()[1]

    @property
    def flavor(self):
        return 'ar'

    def can_link(self, format, langs):
        return format == self.builder.object_format

    @property
    def has_link_macros(self):
        # We only need to define LIBFOO_EXPORTS/LIBFOO_STATIC macros on
        # platforms that have different import/export rules for libraries. We
        # approximate this by checking if the platform uses import libraries,
        # and only define the macros if it does.
        return self.env.target_platform.has_import_library

    def compile_options(self, context):
        options = opts.option_list()
        if self.builder.object_format != 'coff':
            options.append(opts.pic())
        options.extend(self.forwarded_compile_options(context))
        return options

    def forwarded_compile_options(self, context):
        options = opts.option_list()
        if self.has_link_macros:
            options.append(opts.define(library_macro(
                context.name, 'static_library'
            )))
        return options

    def flags(self, options, output=None, mode='normal'):
        flags = []
        for i in options:
            if isinstance(i, safe_str.stringy_types):
                flags.append(i)
            else:
                raise TypeError('unknown option type {!r}'.format(type(i)))
        return flags

    def _call(self, cmd, input, output, flags=None):
        return list(chain(
            cmd, iterate(flags), [output], iterate(input)
        ))

    def output_file(self, name, context):
        head, tail = os.path.split(name)
        path = os.path.join(head, 'lib' + tail + '.a')
        return StaticLibrary(Path(path), self.builder.object_format,
                             context.langs)



from . import builder, cc, msvc
from .common import choose_builder
from ..languages import known_formats, known_langs

with known_langs.make('c') as x:
    x.vars(compiler='CC', flags='CFLAGS')
    x.exts(source=['.c'], header=['.h'])

with known_langs.make('c++') as x:
    x.vars(compiler='CXX', flags='CXXFLAGS')
    x.exts(source=['.cpp', '.cc', '.cp', '.cxx', '.CPP', '.c++', '.C'],
           header=['.hpp', '.hh', '.hp', '.hxx', '.HPP', '.h++', '.H'])
    x.auxexts(header=['.h'])

with known_langs.make('objc') as x:
    x.vars(compiler='OBJC', flags='OBJCFLAGS')
    x.exts(source=['.m'])
    x.auxexts(header=['.h'])

with known_langs.make('objc++') as x:
    x.vars(compiler='OBJCXX', flags='OBJCXXFLAGS')
    x.exts(source=['.mm', '.M'])
    x.auxexts(header=['.h'])

with known_formats.make('native', mode='dynamic') as x:
    x.vars(linker='LD', flags='LDFLAGS', libs='LDLIBS')
with known_formats.make('native', mode='static') as x:
    x.vars(linker='AR', flags='ARFLAGS')

_posix_cmds = {
    'c'     : 'cc' ,
    'c++'   : 'c++',
    'objc'  : 'cc' ,
    'objc++': 'c++',
}
_windows_cmds = {
    'c'     : ['cl', 'clang-cl', 'cc', 'gcc', 'clang'],
    'c++'   : ['cl', 'clang-cl', 'c++', 'g++', 'clang++'],
    'objc'  : ['cc', 'gcc', 'clang'],
    'objc++': ['c++', 'g++', 'clang++'],
}

_builders = (cc.CcBuilder, msvc.MsvcBuilder)


@builder('c', 'c++', 'objc', 'objc++')
def c_family_builder(env, lang):
    cmd_map = (_windows_cmds if env.host_platform.family == 'windows'
               else _posix_cmds)
    return choose_builder(env, known_langs[lang], cmd_map[lang], _builders)



from . import *


class TestExtraDeps(IntegrationTest):
    def __init__(self, *args, **kwargs):
        IntegrationTest.__init__(self, 'extra_deps', *args, **kwargs)

    def test_build(self):
        self.build(executable('program'))
        self.assertOutput([executable('program')], 'hello, world!\n')

        self.assertExists(output_file('1'))
        self.assertExists(output_file('2'))
        self.assertExists(output_file('3'))

    def test_touch_3(self):
        self.build('touch3')

        self.assertNotExists(output_file('1'))
        self.assertExists(output_file('2'))
        self.assertExists(output_file('3'))



import fnmatch
import os
import re
from enum import IntEnum

from . import builtin
from ..file_types import File
from ..iterutils import iterate
from ..backends.make import writer as make
from ..backends.ninja import writer as ninja
from ..backends.make.syntax import Writer, Syntax
from ..build_inputs import build_input
from ..path import exists, isdir, islink, Path, Root
from ..platforms import known_platforms

build_input('find_dirs')(lambda build_inputs, env: set())
depfile_name = '.bfg_find_deps'
exclude_globs = ['.*#', '*~', '#*#']


@builtin.function()
class FindResult(IntEnum):
    include = 0
    not_now = 1
    exclude = 2


def write_depfile(env, path, output, seen_dirs, makeify=False):
    with open(path.string(env.base_dirs), 'w') as f:
        # Since this file is in the build dir, we can use relative dirs for
        # deps also in the build dir.
        roots = env.base_dirs.copy()
        roots[Root.builddir] = None

        out = Writer(f)
        out.write(output.string(roots), Syntax.target)
        out.write_literal(':')
        for i in seen_dirs:
            out.write_literal(' ')
            out.write(i.string(roots), Syntax.dependency)
        out.write_literal('\n')
        if makeify:
            for i in seen_dirs:
                out.write(i.string(roots), Syntax.target)
                out.write_literal(':\n')


def _listdir(path, variables=None):
    dirs, nondirs = [], []
    try:
        names = os.listdir(path.string(variables))
        for name in names:
            curpath = path.append(name)
            if isdir(curpath, variables):
                dirs.append(curpath)
            else:
                nondirs.append(curpath)
    except OSError:
        pass
    return dirs, nondirs


def _walk_flat(top, variables=None):
    if exists(top, variables):
        yield (top,) + _listdir(top, variables)


def _walk_recursive(top, variables=None):
    if not exists(top, variables):
        return
    dirs, nondirs = _listdir(top, variables)
    yield top, dirs, nondirs
    for d in dirs:
        if not islink(d, variables):
            for i in _walk_recursive(d, variables):
                yield i


def _make_filter_from_glob(match_type, matches, extra, exclude):
    matches = [re.compile(fnmatch.translate(i)) for i in iterate(matches)]
    extra = [re.compile(fnmatch.translate(i)) for i in iterate(extra)]
    exclude = [re.compile(fnmatch.translate(i)) for i in iterate(exclude)]

    def fn(path, type):
        name = path.basename()
        if match_type in {type, '*'}:
            if any(ex.match(name) for ex in exclude):
                return FindResult.exclude
            if any(ex.match(name) for ex in matches):
                return FindResult.include
            elif any(ex.match(name) for ex in extra):
                return FindResult.not_now
        return FindResult.exclude
    return fn


@builtin.function('env')
def filter_by_platform(env, path, type):
    my_plat = {env.target_platform.genus, env.target_platform.family}
    sub = '|'.join(re.escape(i) for i in known_platforms if i not in my_plat)
    ex = r'(^|/|_)(' + sub + r')(\.[^\.]+$|$|/)'
    return (FindResult.not_now if re.search(ex, path.suffix)
            else FindResult.include)


def _combine_filters(*args):
    return lambda path, type: max(f(path, type) for f in args)


def _find_files(env, paths, filter, flat, seen_dirs=None):
    # "Does the walker choose the path, or the path the walker?" - Garth Nix
    walker = _walk_flat if flat else _walk_recursive

    for p in paths:
        yield p, 'd', filter(p, 'd')
    for p in paths:
        for base, dirs, files in walker(p, env.base_dirs):
            if seen_dirs is not None:
                seen_dirs.append(base)

            for p in dirs:
                yield p, 'd', filter(p, 'd')
            for p in files:
                yield p, 'f', filter(p, 'f')


def find(env, path='.', name='*', type='*', extra=None, exclude=exclude_globs,
         flat=False):
    glob_filter = _make_filter_from_glob(type, name, extra, exclude)
    paths = [Path.ensure(i, Root.srcdir) for i in iterate(path)]

    results = []
    for path, type, matched in _find_files(env, paths, glob_filter, flat):
        if matched == FindResult.include:
            results.append(path)
    return results


@builtin.function('builtins', 'build_inputs', 'env')
def find_files(builtins, build_inputs, env, path='.', name='*', type='*',
               extra=None, exclude=exclude_globs, filter=None, flat=False,
               file_type=None, dir_type=None, dist=True, cache=True):
    final_filter = _make_filter_from_glob(type, name, extra, exclude)
    if filter:
        final_filter = _combine_filters(final_filter, filter)

    types = {'f': file_type or builtins['auto_file'],
             'd': dir_type or builtins['directory']}
    extra_types = {'f': builtins['generic_file'], 'd': builtins['directory']}

    paths = [i.path if isinstance(i, File) else Path.ensure(i, Root.srcdir)
             for i in iterate(path)]

    found, seen_dirs = [], []
    for path, type, matched in _find_files(env, paths, final_filter, flat,
                                           seen_dirs):
        if matched == FindResult.include:
            found.append(types[type](path, dist=dist))
        elif matched == FindResult.not_now and dist:
            extra_types[type](path, dist=dist)

    if cache:
        build_inputs['find_dirs'].update(seen_dirs)
        build_inputs['regenerate'].depfile = depfile_name
    return found


@builtin.function('builtins')
def find_paths(builtins, *args, **kwargs):
    return [i.path for i in builtins['find_files'](*args, **kwargs)]


@make.post_rule
def make_find_dirs(build_inputs, buildfile, env):
    if build_inputs['find_dirs']:
        write_depfile(env, Path(depfile_name), make.filepath,
                      build_inputs['find_dirs'], makeify=True)
        buildfile.include(depfile_name)


@ninja.post_rule
def ninja_find_dirs(build_inputs, buildfile, env):
    if build_inputs['find_dirs']:
        write_depfile(env, Path(depfile_name), ninja.filepath,
                      build_inputs['find_dirs'])



import os
import re
from itertools import chain

from .common import BuildCommand, Builder, check_which, not_buildroot
from .. import options as opts, safe_str, shell
from ..builtins.file_types import make_immediate_file
from ..exceptions import PackageResolutionError
from ..file_types import *
from ..iterutils import flatten, iterate, uniques
from ..languages import known_formats
from ..packages import Package
from ..path import Path, Root
from ..versioning import detect_version

_warning_flags = {
    'java': {
        opts.WarningValue.disable: '-nowarn',
        opts.WarningValue.error: '-Werror',
        opts.WarningValue.all: '-Xlint:all',
    },
    'scala': {
        opts.WarningValue.disable: '-nowarn',
        opts.WarningValue.error: '-Xfatal-errors',
        opts.WarningValue.all: '-Xlint:_',
    },
}


class JvmBuilder(Builder):
    def __init__(self, env, langinfo, command, version_output):
        # The default command name to run JVM programs is (usually) the same as
        # the name of the language, so we'll just use that here as the default.
        run_name = langinfo.var('runner').lower()
        run_command = check_which(
            env.getvar(langinfo.var('runner'), langinfo.name),
            kind='{} runner'.format(langinfo.name)
        )

        Builder.__init__(self, langinfo.name, *self._parse_brand(
            env, langinfo.name, version_output, run_command
        ))
        self.object_format = 'jvm'

        name = langinfo.var('compiler').lower()
        ldinfo = known_formats['jvm', 'dynamic']

        flags_name = langinfo.var('flags').lower()
        flags = shell.split(env.getvar(langinfo.var('flags'), ''))

        jar_name = ldinfo.var('linker').lower()
        jar_command = check_which(env.getvar(ldinfo.var('linker'), 'jar'),
                                  kind='jar builder')

        jarflags_name = ldinfo.var('flags').lower()
        jarflags = shell.split(env.getvar(ldinfo.var('flags'), 'cfm'))

        self.compiler = JvmCompiler(self, env, name, command, flags_name,
                                    flags)
        self._linker = JarMaker(self, env, jar_name, jar_command,
                                jarflags_name, jarflags)
        self.packages = JvmPackageResolver(self, env, run_command)
        self.runner = JvmRunner(self, env, run_name, run_command)

    @staticmethod
    def _parse_brand(env, lang, version_output, run_command):
        brand = 'unknown'
        version = None
        if lang == 'java':
            try:
                # Get the brand from the run command (rather than the compile
                # command).
                output = env.execute(
                    run_command + ['-version'], stdout=shell.Mode.pipe,
                    stderr=shell.Mode.stdout
                )
                if re.search(r'Java\(TM\) (\w+ )?Runtime Environment', output):
                    brand = 'oracle'
                elif 'OpenJDK Runtime Environment' in output:
                    brand = 'openjdk'
            except (OSError, shell.CalledProcessError):
                pass
            version = detect_version(version_output)
        elif lang == 'scala':
            if 'EPFL' in version_output:
                brand = 'epfl'
                version = detect_version(version_output)
        else:  # pragma: no cover
            raise ValueError('unrecognized language {!r}'.format(lang))

        return brand, version

    @staticmethod
    def check_command(env, command):
        return env.execute(command + ['-version'], stdout=shell.Mode.pipe,
                           stderr=shell.Mode.stdout)

    @property
    def flavor(self):
        return 'jvm'

    @property
    def family(self):
        return 'jvm'

    @property
    def can_dual_link(self):
        return False

    def linker(self, mode):
        if mode == 'static_library':
            raise ValueError('static linking not supported with {}'.format(
                self.brand
            ))
        if mode not in ('executable', 'shared_library'):
            raise KeyError(mode)
        return self._linker


class JvmCompiler(BuildCommand):
    def __init__(self, builder, env, name, command, flags_name, flags):
        BuildCommand.__init__(self, builder, env, name, name, command,
                              flags=(flags_name, flags))

    @property
    def deps_flavor(self):
        return None

    @property
    def needs_libs(self):
        return True

    @property
    def accepts_pch(self):
        return False

    def _call(self, cmd, input, output, flags=None):
        jvmoutput = self.env.tool('jvmoutput')
        result = list(chain(
            cmd, self._always_flags, iterate(flags), [input]
        ))
        return jvmoutput(output, result)

    @property
    def _always_flags(self):
        return ['-verbose', '-d', '.']

    def flags(self, options, output=None, mode='normal'):
        flags, class_path = [], []
        for i in options:
            if isinstance(i, opts.lib):
                class_path.append(i.library.path)
            elif isinstance(i, opts.warning):
                for j in i.value:
                    try:
                        flags.append(_warning_flags[self.lang][j])
                    except KeyError:
                        raise ValueError('unsupported warning level {!r}'
                                         .format(j))
            elif isinstance(i, opts.debug):
                flags.append('-g' if self.lang == 'java' else '-g:vars')
            elif isinstance(i, opts.optimize):
                if ( self.lang == 'scala' and
                     (opts.OptimizeValue.size in i.value or
                      opts.OptimizeValue.speed in i.value) ):
                    flags.append('-optimize')
            elif isinstance(i, safe_str.stringy_types):
                flags.append(i)
            else:
                raise TypeError('unknown option type {!r}'.format(type(i)))

        if class_path:
            flags.extend(['-cp', safe_str.join(uniques(class_path),
                                               os.pathsep)])
        return flags

    def default_name(self, input, context):
        return input.path.stripext().suffix

    def output_file(self, name, context):
        return ObjectFileList(Path(name + '.classlist'), Path(name + '.class'),
                              self.builder.object_format, self.lang)


class JarMaker(BuildCommand):
    def __init__(self, builder, env, name, command, jarflags_name, jarflags):
        BuildCommand.__init__(self, builder, env, 'jar', 'jar', command,
                              flags=(jarflags_name, jarflags))

    @property
    def flavor(self):
        return 'jar'

    def can_link(self, format, langs):
        return format == 'jvm'

    @property
    def needs_libs(self):
        return False

    @property
    def has_link_macros(self):
        return False

    def pre_build(self, build, name, context):
        # Fix up paths for the Class-Path field: escape spaces, use forward
        # slashes on Windows, and prefix Windows drive letters with '/' to
        # disambiguate them from URLs.
        def fix_path(p):
            if self.env.target_platform.family == 'windows':
                if p[1] == ':':
                    p = '/' + p
                p = p.replace('\\', '/')
            return p.replace(' ', '%20')

        libs = getattr(context, 'libs', []) + flatten(
            i.libs for i in getattr(context, 'packages', [])
        )
        dirs = uniques(i.path for i in libs)
        base = Path(name).parent()

        context.manifest = File(Path(name + '-manifest.txt'))
        with make_immediate_file(build, self.env, context.manifest) as out:
            classpath = ' '.join(fix_path(i.relpath(base)) for i in dirs)
            if classpath:
                out.write('Class-Path: {}\n'.format(classpath))

            if getattr(context, 'entry_point', None):
                out.write('Main-Class: {}\n'.format(context.entry_point))

        return opts.option_list()

    def _call(self, cmd, input, output, manifest, flags=None):
        return list(chain(
            cmd, iterate(flags), [output, manifest], iterate(input)
        ))

    def transform_input(self, input):
        return ['@' + safe_str.safe_str(i) if isinstance(i, ObjectFileList)
                else i for i in input]

    def flags(self, options, output=None, mode='normal'):
        flags = []
        for i in options:
            if isinstance(i, safe_str.stringy_types):
                flags.append(i)
            elif isinstance(i, opts.debug):
                pass
            elif isinstance(i, opts.optimize):
                pass
            else:
                raise TypeError('unknown option type {!r}'.format(type(i)))
        return flags

    def output_file(self, name, context):
        if getattr(context, 'entry_point', None):
            filetype = ExecutableLibrary
        else:
            filetype = Library
        return filetype(Path(name + '.jar'), self.builder.object_format,
                        self.lang)


class JvmPackage(Package):
    def __init__(self, name, format, libs=None):
        Package.__init__(self, name, format)
        self.libs = libs or []

    def compile_options(self, compiler):
        return opts.option_list(opts.lib(i) for i in self.libs)

    def link_options(self, linker):
        return opts.option_list()


class JvmPackageResolver(object):
    def __init__(self, builder, env, command):
        self.builder = builder

        if self.lang == 'scala':
            extra_env = {'JAVA_OPTS': '-XshowSettings:properties'}
            args = ['-version']
            returncode = 1
        else:
            extra_env = None
            args = ['-XshowSettings:properties', '-version']
            returncode = 0

        try:
            output = env.execute(
                command + args, env=extra_env, stdout=shell.Mode.devnull,
                stderr=shell.Mode.pipe, returncode=returncode
            )
            self.ext_dirs = self._get_dirs('java.ext.dirs', output)
            self.classpath = self._get_dirs('java.class.path', output)
        except (OSError, shell.CalledProcessError):
            self.ext_dirs = []
            self.classpath = []

    @property
    def lang(self):
        return self.builder.lang

    def _get_dirs(self, key, output):
        ex = r'^(\s*){} = (.*(?:\n\1\s+.*)*)'.format(re.escape(key))
        m = re.search(ex, output, re.MULTILINE)
        if not m:
            return []
        return [i.strip() for i in m.group(2).split('\n')]

    def _library(self, name):
        jarname = name + '.jar'
        for base in self.ext_dirs:
            fullpath = os.path.join(base, jarname)
            if os.path.exists(fullpath):
                return Library(Path(fullpath, Root.absolute),
                               self.builder.object_format)

        for path in self.classpath:
            if os.path.basename(path) == jarname and os.path.exists(path):
                return Library(Path(path, Root.absolute),
                               self.builder.object_format)

        raise PackageResolutionError("unable to find library '{}'"
                                     .format(name))

    def resolve(self, name, version, kind, headers, libs):
        return JvmPackage(name, self.builder.object_format,
                          libs=[self._library(name)])


class JvmRunner(BuildCommand):
    def __init__(self, builder, env, name, command):
        BuildCommand.__init__(self, builder, env, name, name, command)

    def _call(self, cmd, file, cp=None, jar=False):
        result = list(cmd)
        if jar and self.lang != 'scala':
            result.append('-jar')
        if not_buildroot(cp):
            result.extend(['-cp', cp])
        result.append(file)
        return result

    def run_arguments(self, file):
        if isinstance(file, Executable):
            return self(file, jar=True)
        elif isinstance(file, ObjectFileList):
            return self.run_arguments(file.object_file)
        elif isinstance(file, ObjectFile):
            return self(file.path.stripext().basename(), cp=file.path.parent())
        raise TypeError('expected an executable or object file for {} to run'
                        .format(self.lang))



import os
import subprocess
from enum import Enum

from .list import shell_list  # noqa
from ..iterutils import listify
from ..path import Path
from ..platforms import platform_name
from ..safe_str import safe_str

windows_names = ('winnt', 'win9x', 'msdos')

if platform_name() in windows_names:
    from .windows import *  # noqa
else:
    from .posix import *  # noqa

Mode = Enum('Mode', ['normal', 'pipe', 'stdout', 'devnull'])
CalledProcessError = subprocess.CalledProcessError


def which(names, env=os.environ, base_dirs=None, resolve=False,
          kind='executable'):
    paths = env.get('PATH', os.defpath).split(os.pathsep)
    exts = ['']
    if platform_name() in windows_names + ('cygwin',):
        exts.extend(env.get('PATHEXT', '').split(os.pathsep))

    names = listify(names)
    if len(names) == 0:
        raise TypeError('must supply at least one name')

    for name in names:
        name = listify(name)
        check = (name[0].string(base_dirs) if isinstance(name[0], Path)
                 else name[0])
        if os.path.isabs(check):
            fullpaths = [check]
        else:
            search = ['.'] if os.path.dirname(check) else paths
            fullpaths = [os.path.normpath(os.path.join(path, check))
                         for path in search]

        for fullpath in fullpaths:
            for ext in exts:
                withext = fullpath + ext
                if os.path.exists(withext):
                    return [withext] + name[1:] if resolve else name

    raise IOError('unable to find {kind}{filler} {names}'.format(
        kind=kind, filler='; tried' if len(names) > 1 else '',
        names=', '.join('{!r}'.format(i) for i in names)
    ))


def execute(args, shell=False, env=None, base_dirs=None, stdout=Mode.normal,
            stderr=Mode.normal, returncode=0):
    def stringify(s):
        s = safe_str(s)
        return s.string(base_dirs) if isinstance(s, Path) else s

    if not shell:
        args = [stringify(i) for i in args]

    devnull = (open(os.devnull, 'wb') if Mode.devnull in (stdout, stderr)
               else None)

    def conv(mode):
        return ({Mode.normal : None,
                 Mode.pipe   : subprocess.PIPE,
                 Mode.stdout : subprocess.STDOUT,
                 Mode.devnull: devnull}).get(mode, mode)

    try:
        proc = subprocess.Popen(
            args, universal_newlines=True, shell=shell, env=env,
            stdout=conv(stdout), stderr=conv(stderr)
        )
        output = proc.communicate()
        if not (returncode == 'any' or
                (returncode == 'fail' and proc.returncode != 0) or
                proc.returncode in listify(returncode)):
            raise CalledProcessError(proc.returncode, args)

        if stdout == Mode.pipe:
            if stderr == Mode.pipe:
                return output
            return output[0]
        return output[1]
    finally:
        if devnull:
            devnull.close()



import importlib
import pkgutil
import warnings

from . import builtin
from .. import exceptions, log, path, safe_str
from ..objutils import memoize


@memoize
def init():
    # Import all the packages in this directory so their hooks get run.
    for _, name, _ in pkgutil.walk_packages(__path__, '.'):
        importlib.import_module(name, __package__)


@builtin.getter('env', name='env', context=('build', 'options'))
def getenv(env):
    return env


@builtin.function(context='*')
def warning(msg):
    warnings.warn(msg)


@builtin.function(context='*')
def info(msg, show_stack=False):
    log.log_stack(log.INFO, msg, show_stack=show_stack, stacklevel=1)


@builtin.function(context='*')
def debug(msg, show_stack=True):
    log.log_stack(log.DEBUG, msg, show_stack=show_stack, stacklevel=1)


for i in dir(exceptions):
    i = getattr(exceptions, i)
    if isinstance(i, type):
        builtin.function(context='*')(i)

for i in (path.Root, path.InstallRoot, safe_str.safe_str,
          safe_str.safe_format):
    builtin.function(context='*')(i)
builtin.function(context='*', name='Path')(path.Path)



from six import iteritems

from . import builtin
from .. import platforms, shell
from ..iterutils import first, isiterable
from ..languages import known_formats, known_langs
from ..path import Path, Root, InstallRoot
from ..shell import posix as pshell

_unsafe_builtins = ['file', '__import__', 'input', 'open', 'raw_input',
                    'reload']


@builtin.getter(name='__builtins__', context='toolchain')
def builtins():
    return {k: v for k, v in iteritems(__builtins__)
            if k not in _unsafe_builtins}


@builtin.getter('env', context='toolchain')
def environ(env):
    return env.variables


@builtin.getter('env', context='toolchain')
def srcdir(env):
    # Make a copy of the srcdir object, since it should be read-only.
    return env.srcdir.reroot(env.srcdir.root)


@builtin.function('env', context='toolchain')
def target_platform(env, platform=None, arch=None):
    env.target_platform = platforms.target.platform_info(platform, arch)


@builtin.function(context='toolchain')
def which(names, resolve=False, strict=True, kind='executable'):
    try:
        return ' '.join(shell.which(names, resolve=resolve, kind=kind))
    except IOError:
        if strict:
            raise
        result = first(names)
        return pshell.join(result) if isiterable(result) else result


@builtin.function('env', context='toolchain')
def compiler(env, names, lang, strict=False):
    var = known_langs[lang].var('compiler')
    env.variables[var] = which(names, strict=strict, kind='compiler')


@builtin.function('env', context='toolchain')
def compile_options(env, options, lang):
    # This only supports strings (and lists of strings) for options, *not*
    # semantic options. It would be nice if we could support semantic options,
    # but we'd either need to know the flavor of compiler at this point (we
    # don't) or we'd have to store the options in some way other than as an
    # environment variable.
    if isiterable(options):
        options = pshell.join(options)
    env.variables[known_langs[lang].var('flags')] = options


@builtin.function('env', context='toolchain')
def runner(env, names, lang, strict=False):
    var = known_langs[lang].var('runner')
    env.variables[var] = which(names, strict=strict, kind='runner')


@builtin.function('env', context='toolchain')
def linker(env, names, format='native', mode='dynamic', strict=False):
    var = known_formats[format, mode].var('linker')
    env.variables[var] = which(names, strict=strict, kind='linker')


@builtin.function('env', context='toolchain')
def link_options(env, options, format='native', mode='dynamic'):
    # As above, this only supports strings (and lists of strings) for options,
    # *not* semantic options.
    if isiterable(options):
        options = pshell.join(options)
    env.variables[known_formats[format, mode].var('flags')] = options


@builtin.function('env', context='toolchain')
def lib_options(env, options, format='native', mode='dynamic'):
    # As above, this only supports strings (and lists of strings) for options,
    # *not* semantic options.
    if isiterable(options):
        options = pshell.join(options)
    env.variables[known_formats[format, mode].var('libs')] = options


@builtin.function('env', 'reload', context='toolchain')
def install_dirs(env, reload, **kwargs):
    if reload:
        return
    for k, v in iteritems(kwargs):
        env.install_dirs[InstallRoot[k]] = Path.ensure(v, Root.absolute)



import platform
import re
import subprocess
from collections import namedtuple
from pkg_resources import get_entry_info

from ..objutils import memoize
from ..versioning import SpecifierSet, Version

__all__ = ['known_platforms', 'parse_triplet', 'Platform', 'PlatformTriplet',
           'platform_name', 'platform_tuple']

# This lists the known platform families and genera.
known_platforms = ['posix', 'windows', 'linux', 'darwin', 'cygwin', 'winnt',
                   'win9x', 'msdos']
PlatformTriplet = namedtuple('PlatformTriplet', ['arch', 'vendor', 'sys',
                                                 'abi'])

_platform_genus = {
    'android': 'linux',
    'ios': 'darwin',
    'macos': 'darwin',
}

_triplet_abi = {'android', 'eabi', 'elf', 'gnu', 'macho'}


def parse_triplet(s, default_vendor='unknown'):
    def _result(arch, vendor_sys, abi=None):
        if len(vendor_sys) == 0:
            raise ValueError('expected <sys>')
        elif len(vendor_sys) == 1:
            return PlatformTriplet(arch, default_vendor, vendor_sys[0], abi)
        elif len(vendor_sys) == 2:
            return PlatformTriplet(arch, vendor_sys[0], vendor_sys[1], abi)
        raise ValueError('too many values')

    bits = s.split('-')
    if len(bits) < 2:
        raise ValueError('expected at least <arch>-<sys>')

    if bits[-1] in _triplet_abi:
        return _result(bits[0], bits[1:-1], bits[-1])
    return _result(bits[0], bits[1:], None)


@memoize
def platform_name():
    system = platform.system().lower()
    if system.startswith('cygwin'):
        return 'cygwin'

    if system == 'windows':
        try:
            uname = subprocess.check_output(
                'uname', universal_newlines=True
            ).lower()
            if uname.startswith('cygwin'):
                return 'cygwin'
        except OSError:
            pass

        version = Version(platform.version())
        if version not in SpecifierSet('!=3.10.528,!=3.50.807,!=3.51.1057,' +
                                       '!=4.00.1381,<5'):
            return 'winnt'
        elif version in SpecifierSet('>=4'):
            return 'win9x'
        return 'msdos'
    elif system == 'linux':
        try:
            distro = subprocess.check_output(
                ['lsb_release', '-is'], universal_newlines=True
            ).lower()
            if distro == 'android':
                return 'android'
        except OSError:
            pass
        return system
    elif system == 'darwin':
        machine = platform.machine()
        if re.match(r'(iPhone|iPad|iPod)', machine):
            return 'ios'
        return 'macos'

    # Not sure what this is...
    return system


def platform_tuple(name=None):
    if name is None:
        name = platform_name()
    return _platform_genus.get(name, str(name)), name


class Platform(object):
    def __init__(self, genus, species, arch):
        self.genus = genus
        self.species = species
        self.arch = arch

    @property
    def name(self):
        return self.species

    @property
    def _triplet_vendor(self):
        return 'pc' if re.match(r'i.86$', self.arch) else 'unknown'

    @property
    def triplet(self):
        return '{}-{}-{}'.format(self.arch, self._triplet_vendor,
                                 self._triplet_sys_abi)

    def __repr__(self):
        return '<{}({})>'.format(type(self).__name__, self.triplet)

    def to_json(self):
        return {
            'genus': self.genus,
            'species': self.species,
            'arch': self.arch,
        }

    def __eq__(self, rhs):
        return (self.genus == rhs.genus and self.species == rhs.species and
                self.arch == rhs.arch)

    def __ne__(self, rhs):
        return not self == rhs


@memoize
def _get_platform_info(kind, genus, species, arch):
    entry_point = 'bfg9000.platforms.{}'.format(kind)
    entry = get_entry_info('bfg9000', entry_point, genus)
    if entry is None:
        # Fall back to a generic POSIX system if we don't recognize the
        # platform name.
        entry = get_entry_info('bfg9000', entry_point, 'posix')
    return entry.load()(genus, species, arch)


def _platform_info(kind, name=None, arch=None):
    if name is None:
        name = platform_name()
    if arch is None:
        arch = platform.machine()
    genus, species = platform_tuple(name)
    return _get_platform_info(kind, genus, species, arch)



from collections import Iterable
from six import iteritems, string_types

__all__ = ['default_sentinel', 'first', 'flatten', 'isiterable', 'iterate',
           'listify', 'merge_dicts', 'merge_into_dict', 'recursive_walk',
           'slice_dict', 'tween', 'uniques', 'unlistify']

# This could go in a funcutils module if we ever create one...
default_sentinel = object()


def isiterable(thing):
    return isinstance(thing, Iterable) and not isinstance(thing, string_types)


def iterate(thing):
    def generate_none():
        return iter(())

    def generate_one(x):
        yield x

    if thing is None:
        return generate_none()
    elif isiterable(thing):
        return iter(thing)
    else:
        return generate_one(thing)


def listify(thing, always_copy=False, scalar_ok=True, type=list):
    if not always_copy and isinstance(thing, type):
        return thing
    if scalar_ok:
        thing = iterate(thing)
    elif not isiterable(thing):
        raise TypeError('expected an iterable')
    return type(thing)


def first(thing, default=default_sentinel):
    try:
        return next(iterate(thing))
    except StopIteration:
        if default is not default_sentinel:
            return default
        raise LookupError()


def unlistify(thing):
    if len(thing) == 0:
        return None
    elif len(thing) == 1:
        return thing[0]
    else:
        return thing


def flatten(iterables, type=list):
    result = type()
    # Performance tests on the kinds of data common in bfg (generators calling
    # a function returning a list on each step) show that this is consistently
    # faster than using `list(chain.from_iterable(iterables))`. It's also
    # *much* faster than using `sum(iterables, [])`.
    for i in iterables:
        result.extend(i)
    return result


def tween(iterable, delim, prefix=None, suffix=None):
    first = True
    for i in iterable:
        if first:
            first = False
            if prefix is not None:
                yield prefix
        else:
            yield delim
        yield i
    if not first and suffix is not None:
        yield suffix


def uniques(iterable):
    def generate_uniques(iterable):
        seen = set()
        for item in iterable:
            if item not in seen:
                seen.add(item)
                yield item
    return list(generate_uniques(iterable))


def recursive_walk(thing, attr, children_attr=None):
    for i in getattr(thing, attr):
        yield i
    for i in getattr(thing, children_attr or attr):
        for j in recursive_walk(i, attr, children_attr):
            yield j


def merge_into_dict(dst, *args):
    for d in args:
        for k, v in iteritems(d):
            curr = dst.get(k)
            if isinstance(v, dict):
                if curr is None:
                    dst[k] = dict(v)
                elif isinstance(curr, dict):
                    merge_into_dict(curr, v)
                else:
                    raise TypeError('type mismatch for {}'.format(k))
            elif isiterable(v):
                if curr is None:
                    dst[k] = type(v)(v)
                elif isiterable(curr):
                    curr.extend(v)
                else:
                    raise TypeError('type mismatch for {}'.format(k))
            elif v is not None:
                if curr is not None and isiterable(curr):
                    raise TypeError('type mismatch for {}'.format(k))
                dst[k] = v
            elif k not in dst:
                dst[k] = None  # v is always None here


def merge_dicts(*args):
    dst = {}
    merge_into_dict(dst, *args)
    return dst


def slice_dict(d, keys):
    result = {}
    for k in keys:
        if k in d:
            result[k] = d.pop(k)
    return result
